import os
import json
from subprocess import CalledProcessError
import shutil
import tempfile
import types
from mock import call, MagicMock, mock_open, patch, sentinel
from testtools import TestCase
import yaml

import six
import io

from charmhelpers.core import hookenv

if six.PY3:
    import pickle
else:
    import cPickle as pickle


CHARM_METADATA = b"""name: testmock
summary: test mock summary
description: test mock description
requires:
    testreqs:
        interface: mock
provides:
    testprov:
        interface: mock
peers:
    testpeer:
        interface: mock
"""


def _clean_globals():
    hookenv.cache.clear()
    del hookenv._atstart[:]
    del hookenv._atexit[:]


class ConfigTest(TestCase):
    def setUp(self):
        super(ConfigTest, self).setUp()

        _clean_globals()
        self.addCleanup(_clean_globals)

        self.charm_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.charm_dir))

        patcher = patch.object(hookenv, 'charm_dir', lambda: self.charm_dir)
        self.addCleanup(patcher.stop)
        patcher.start()

    def test_init(self):
        d = dict(foo='bar')
        c = hookenv.Config(d)

        self.assertEqual(c['foo'], 'bar')
        self.assertEqual(c._prev_dict, None)

    def test_init_empty_state_file(self):
        d = dict(foo='bar')
        c = hookenv.Config(d)

        with open(c.path, 'w') as f:
            f.close()

        self.assertEqual(c['foo'], 'bar')
        self.assertEqual(c._prev_dict, None)

    def test_init_invalid_state_file(self):
        d = dict(foo='bar')
        c = hookenv.Config(d)

        with open(c.path, 'w') as f:
            f.write('blah')

        self.assertEqual(c['foo'], 'bar')
        self.assertEqual(c._prev_dict, None)

    def test_load_previous(self):
        d = dict(foo='bar')
        c = hookenv.Config()

        with open(c.path, 'w') as f:
            json.dump(d, f)

        c.load_previous()
        self.assertEqual(c._prev_dict, d)

    def test_load_previous_alternate_path(self):
        d = dict(foo='bar')
        c = hookenv.Config()

        alt_path = os.path.join(self.charm_dir, '.alt-config')
        with open(alt_path, 'w') as f:
            json.dump(d, f)

        c.load_previous(path=alt_path)
        self.assertEqual(c._prev_dict, d)
        self.assertEqual(c.path, alt_path)

    def test_changed_without_prev_dict(self):
        d = dict(foo='bar')
        c = hookenv.Config(d)

        self.assertTrue(c.changed('foo'))

    def test_changed_with_prev_dict(self):
        c = hookenv.Config(dict(foo='bar', a='b'))
        c.save()
        c = hookenv.Config(dict(foo='baz', a='b'))

        self.assertTrue(c.changed('foo'))
        self.assertFalse(c.changed('a'))

    def test_previous_without_prev_dict(self):
        c = hookenv.Config()

        self.assertEqual(c.previous('foo'), None)

    def test_previous_with_prev_dict(self):
        c = hookenv.Config(dict(foo='bar'))
        c.save()
        c = hookenv.Config(dict(foo='baz', a='b'))

        self.assertEqual(c.previous('foo'), 'bar')
        self.assertEqual(c.previous('a'), None)

    def test_save_without_prev_dict(self):
        c = hookenv.Config(dict(foo='bar'))
        c.save()

        with open(c.path, 'r') as f:
            self.assertEqual(c, json.load(f))
            self.assertEqual(c, dict(foo='bar'))
            self.assertEqual(os.stat(c.path).st_mode & 0o777, 0o600)

    def test_save_with_prev_dict(self):
        c = hookenv.Config(dict(foo='bar'))
        c.save()
        c = hookenv.Config(dict(a='b'))
        c.save()

        with open(c.path, 'r') as f:
            self.assertEqual(c, json.load(f))
            self.assertEqual(c, dict(foo='bar', a='b'))
            self.assertEqual(os.stat(c.path).st_mode & 0o777, 0o600)

    def test_deep_change(self):
        # After loading stored data into our previous dictionary,
        # it gets copied into our current dictionary. If this is not
        # a deep copy, then mappings and lists will be shared instances
        # and changes will not be detected.
        c = hookenv.Config(dict(ll=[]))
        c.save()
        c = hookenv.Config()
        c['ll'].append(42)
        self.assertTrue(c.changed('ll'), 'load_previous() did not deepcopy')

    def test_getitem(self):
        c = hookenv.Config(dict(foo='bar'))
        c.save()
        c = hookenv.Config(dict(baz='bam'))

        self.assertRaises(KeyError, lambda: c['missing'])
        self.assertEqual(c['foo'], 'bar')
        self.assertEqual(c['baz'], 'bam')

    def test_get(self):
        c = hookenv.Config(dict(foo='bar'))
        c.save()
        c = hookenv.Config(dict(baz='bam'))

        self.assertIsNone(c.get('missing'))
        self.assertIs(c.get('missing', sentinel.missing), sentinel.missing)
        self.assertEqual(c.get('foo'), 'bar')
        self.assertEqual(c.get('baz'), 'bam')

    def test_keys(self):
        c = hookenv.Config(dict(foo='bar'))
        c["baz"] = "bar"
        self.assertEqual(sorted([six.u("foo"), "baz"]), sorted(c.keys()))

    def test_in(self):
        # Test behavior of the in operator.

        prev_path = os.path.join(hookenv.charm_dir(),
                                 hookenv.Config.CONFIG_FILE_NAME)
        with open(prev_path, 'w') as f:
            json.dump(dict(user='one'), f)
        c = hookenv.Config(dict(charm='one'))

        # Items that exist in the dict exist. Items that don't don't.
        self.assertTrue('user' in c)
        self.assertTrue('charm' in c)
        self.assertFalse('bar' in c)

        # Adding items works as expected.
        c['user'] = 'two'
        c['charm'] = 'two'
        c['bar'] = 'two'
        self.assertTrue('user' in c)
        self.assertTrue('charm' in c)
        self.assertTrue('bar' in c)
        c.save()
        self.assertTrue('user' in c)
        self.assertTrue('charm' in c)
        self.assertTrue('bar' in c)

        # Removing items works as expected.
        del c['user']
        del c['charm']
        self.assertTrue('user' not in c)
        self.assertTrue('charm' not in c)
        c.save()
        self.assertTrue('user' not in c)
        self.assertTrue('charm' not in c)


class SerializableTest(TestCase):
    def test_serializes_object_to_json(self):
        foo = {
            'bar': 'baz',
        }
        wrapped = hookenv.Serializable(foo)
        self.assertEqual(wrapped.json(), json.dumps(foo))

    def test_serializes_object_to_yaml(self):
        foo = {
            'bar': 'baz',
        }
        wrapped = hookenv.Serializable(foo)
        self.assertEqual(wrapped.yaml(), yaml.dump(foo))

    def test_gets_attribute_from_inner_object_as_dict(self):
        foo = {
            'bar': 'baz',
        }
        wrapped = hookenv.Serializable(foo)

        self.assertEqual(wrapped.bar, 'baz')

    def test_raises_error_from_inner_object_as_dict(self):
        foo = {
            'bar': 'baz',
        }
        wrapped = hookenv.Serializable(foo)

        self.assertRaises(AttributeError, getattr, wrapped, 'baz')

    def test_dict_methods_from_inner_object(self):
        foo = {
            'bar': 'baz',
        }
        wrapped = hookenv.Serializable(foo)
        for meth in ('keys', 'values', 'items'):
            self.assertEqual(sorted(list(getattr(wrapped, meth)())),
                             sorted(list(getattr(foo, meth)())))

        self.assertEqual(wrapped.get('bar'), foo.get('bar'))
        self.assertEqual(wrapped.get('baz', 42), foo.get('baz', 42))
        self.assertIn('bar', wrapped)

    def test_get_gets_from_inner_object(self):
        foo = {
            'bar': 'baz',
        }
        wrapped = hookenv.Serializable(foo)

        self.assertEqual(wrapped.get('foo'), None)
        self.assertEqual(wrapped.get('bar'), 'baz')
        self.assertEqual(wrapped.get('zoo', 'bla'), 'bla')

    def test_gets_inner_object(self):
        foo = {
            'bar': 'baz',
        }
        wrapped = hookenv.Serializable(foo)

        self.assertIs(wrapped.data, foo)

    def test_pickle(self):
        foo = {'bar': 'baz'}
        wrapped = hookenv.Serializable(foo)
        pickled = pickle.dumps(wrapped)
        unpickled = pickle.loads(pickled)

        self.assert_(isinstance(unpickled, hookenv.Serializable))
        self.assertEqual(unpickled, foo)

    def test_boolean(self):
        true_dict = {'foo': 'bar'}
        false_dict = {}

        self.assertIs(bool(hookenv.Serializable(true_dict)), True)
        self.assertIs(bool(hookenv.Serializable(false_dict)), False)

    def test_equality(self):
        foo = {'bar': 'baz'}
        bar = {'baz': 'bar'}
        wrapped_foo = hookenv.Serializable(foo)

        self.assertEqual(wrapped_foo, foo)
        self.assertEqual(wrapped_foo, wrapped_foo)
        self.assertNotEqual(wrapped_foo, bar)


class HelpersTest(TestCase):
    def setUp(self):
        super(HelpersTest, self).setUp()
        _clean_globals()
        self.addCleanup(_clean_globals)

    @patch('subprocess.call')
    def test_logs_messages_to_juju_with_default_level(self, mock_call):
        hookenv.log('foo')

        mock_call.assert_called_with(['juju-log', 'foo'])

    @patch('subprocess.call')
    def test_logs_messages_object(self, mock_call):
        hookenv.log(object)
        mock_call.assert_called_with(['juju-log', repr(object)])

    @patch('subprocess.call')
    def test_logs_messages_with_alternative_levels(self, mock_call):
        alternative_levels = [
            hookenv.CRITICAL,
            hookenv.ERROR,
            hookenv.WARNING,
            hookenv.INFO,
        ]

        for level in alternative_levels:
            hookenv.log('foo', level)
            mock_call.assert_called_with(['juju-log', '-l', level, 'foo'])

    @patch('charmhelpers.core.hookenv._cache_config', None)
    @patch('charmhelpers.core.hookenv.charm_dir')
    @patch('subprocess.check_output')
    def test_gets_charm_config_with_scope(self, check_output, charm_dir):
        check_output.return_value = json.dumps(dict(baz='bar')).encode('UTF-8')
        charm_dir.return_value = '/nonexistent'

        result = hookenv.config(scope='baz')

        self.assertEqual(result, 'bar')
        check_output.assert_called_with(['config-get', '--all',
                                         '--format=json'])

        # The result can be used like a string
        self.assertEqual(result[1], 'a')

        # ... because the result is actually a string
        self.assert_(isinstance(result, six.string_types))

    @patch('charmhelpers.core.hookenv.log', lambda *args, **kwargs: None)
    @patch('charmhelpers.core.hookenv._cache_config', None)
    @patch('subprocess.check_output')
    def test_gets_missing_charm_config_with_scope(self, check_output):
        check_output.return_value = b''

        result = hookenv.config(scope='baz')

        self.assertEqual(result, None)
        check_output.assert_called_with(['config-get', '--all',
                                         '--format=json'])

    @patch('charmhelpers.core.hookenv._cache_config', None)
    @patch('charmhelpers.core.hookenv.charm_dir')
    @patch('subprocess.check_output')
    def test_gets_config_without_scope(self, check_output, charm_dir):
        check_output.return_value = json.dumps(dict(baz='bar')).encode('UTF-8')
        charm_dir.return_value = '/nonexistent'

        result = hookenv.config()

        self.assertIsInstance(result, hookenv.Config)
        self.assertEqual(result['baz'], 'bar')
        check_output.assert_called_with(['config-get', '--all',
                                         '--format=json'])

    @patch('charmhelpers.core.hookenv.log')
    @patch('charmhelpers.core.hookenv._cache_config', None)
    @patch('charmhelpers.core.hookenv.charm_dir')
    @patch('subprocess.check_output')
    def test_gets_charm_config_invalid_json_with_scope(self,
                                                       check_output,
                                                       charm_dir,
                                                       log):
        check_output.return_value = '{"invalid: "json"}'.encode('UTF-8')
        charm_dir.return_value = '/nonexistent'

        result = hookenv.config(scope='invalid')

        self.assertEqual(result, None)
        cmd_line = ['config-get', '--all', '--format=json']
        check_output.assert_called_with(cmd_line)
        log.assert_called_with(
            'Unable to parse output from config-get: '
            'config_cmd_line="{}" message="{}"'
            .format(str(cmd_line),
                    "Expecting ':' delimiter: line 1 column 13 (char 12)"),
            level=hookenv.ERROR,
        )

    @patch('charmhelpers.core.hookenv.log')
    @patch('charmhelpers.core.hookenv._cache_config', None)
    @patch('charmhelpers.core.hookenv.charm_dir')
    @patch('subprocess.check_output')
    def test_gets_charm_config_invalid_utf8_with_scope(self,
                                                       check_output,
                                                       charm_dir,
                                                       log):
        check_output.return_value = b'{"invalid: "json"}\x9D'
        charm_dir.return_value = '/nonexistent'

        result = hookenv.config(scope='invalid')

        self.assertEqual(result, None)
        cmd_line = ['config-get', '--all', '--format=json']
        check_output.assert_called_with(cmd_line)
        try:
            # Python3
            log.assert_called_with(
                'Unable to parse output from config-get: '
                'config_cmd_line="{}" message="{}"'
                .format(str(cmd_line),
                        "'utf8' codec can't decode byte 0x9d in position "
                        "18: invalid start byte"),
                level=hookenv.ERROR,
            )
        except AssertionError:
            # Python2.7
            log.assert_called_with(
                'Unable to parse output from config-get: '
                'config_cmd_line="{}" message="{}"'
                .format(str(cmd_line),
                        "'utf-8' codec can't decode byte 0x9d in position "
                        "18: invalid start byte"),
                level=hookenv.ERROR,
            )

    @patch('charmhelpers.core.hookenv._cache_config', {'baz': 'bar'})
    @patch('charmhelpers.core.hookenv.charm_dir')
    @patch('subprocess.check_output')
    def test_gets_config_from_cache_without_scope(self,
                                                  check_output,
                                                  charm_dir):
        charm_dir.return_value = '/nonexistent'

        result = hookenv.config()

        self.assertEqual(result['baz'], 'bar')
        self.assertFalse(check_output.called)

    @patch('charmhelpers.core.hookenv._cache_config', {'baz': 'bar'})
    @patch('charmhelpers.core.hookenv.charm_dir')
    @patch('subprocess.check_output')
    def test_gets_config_from_cache_with_scope(self,
                                               check_output,
                                               charm_dir):
        charm_dir.return_value = '/nonexistent'

        result = hookenv.config('baz')

        self.assertEqual(result, 'bar')

        # The result can be used like a string
        self.assertEqual(result[1], 'a')

        # ... because the result is actually a string
        self.assert_(isinstance(result, six.string_types))

        self.assertFalse(check_output.called)

    @patch('charmhelpers.core.hookenv._cache_config', {'foo': 'bar'})
    @patch('subprocess.check_output')
    def test_gets_missing_charm_config_from_cache_with_scope(self,
                                                             check_output):

        result = hookenv.config(scope='baz')

        self.assertEqual(result, None)
        self.assertFalse(check_output.called)

    @patch('charmhelpers.core.hookenv.os')
    def test_gets_the_local_unit(self, os_):
        os_.environ = {
            'JUJU_UNIT_NAME': 'foo',
        }

        self.assertEqual(hookenv.local_unit(), 'foo')

    @patch('charmhelpers.core.hookenv.unit_get')
    def test_gets_unit_public_ip(self, _unitget):
        _unitget.return_value = sentinel.public_ip
        self.assertEqual(sentinel.public_ip, hookenv.unit_public_ip())
        _unitget.assert_called_once_with('public-address')

    @patch('charmhelpers.core.hookenv.unit_get')
    def test_gets_unit_private_ip(self, _unitget):
        _unitget.return_value = sentinel.private_ip
        self.assertEqual(sentinel.private_ip, hookenv.unit_private_ip())
        _unitget.assert_called_once_with('private-address')

    @patch('charmhelpers.core.hookenv.os')
    def test_checks_that_is_running_in_relation_hook(self, os_):
        os_.environ = {
            'JUJU_RELATION': 'foo',
        }

        self.assertTrue(hookenv.in_relation_hook())

    @patch('charmhelpers.core.hookenv.os')
    def test_checks_that_is_not_running_in_relation_hook(self, os_):
        os_.environ = {
            'bar': 'foo',
        }

        self.assertFalse(hookenv.in_relation_hook())

    @patch('charmhelpers.core.hookenv.os')
    def test_gets_the_relation_type(self, os_):
        os_.environ = {
            'JUJU_RELATION': 'foo',
        }

        self.assertEqual(hookenv.relation_type(), 'foo')

    @patch('charmhelpers.core.hookenv.os')
    def test_relation_type_none_if_not_in_environment(self, os_):
        os_.environ = {}
        self.assertEqual(hookenv.relation_type(), None)

    @patch('subprocess.check_output')
    @patch('charmhelpers.core.hookenv.relation_type')
    def test_gets_relation_ids(self, relation_type, check_output):
        ids = [1, 2, 3]
        check_output.return_value = json.dumps(ids).encode('UTF-8')
        reltype = 'foo'
        relation_type.return_value = reltype

        result = hookenv.relation_ids()

        self.assertEqual(result, ids)
        check_output.assert_called_with(['relation-ids', '--format=json',
                                         reltype])

    @patch('subprocess.check_output')
    @patch('charmhelpers.core.hookenv.relation_type')
    def test_gets_relation_ids_empty_array(self, relation_type, check_output):
        ids = []
        check_output.return_value = json.dumps(None).encode('UTF-8')
        reltype = 'foo'
        relation_type.return_value = reltype

        result = hookenv.relation_ids()

        self.assertEqual(result, ids)
        check_output.assert_called_with(['relation-ids', '--format=json',
                                         reltype])

    @patch('subprocess.check_output')
    @patch('charmhelpers.core.hookenv.relation_type')
    def test_relation_ids_no_relation_type(self, relation_type, check_output):
        ids = [1, 2, 3]
        check_output.return_value = json.dumps(ids).encode('UTF-8')
        relation_type.return_value = None

        result = hookenv.relation_ids()

        self.assertEqual(result, [])

    @patch('subprocess.check_output')
    @patch('charmhelpers.core.hookenv.relation_type')
    def test_gets_relation_ids_for_type(self, relation_type, check_output):
        ids = [1, 2, 3]
        check_output.return_value = json.dumps(ids).encode('UTF-8')
        reltype = 'foo'

        result = hookenv.relation_ids(reltype)

        self.assertEqual(result, ids)
        check_output.assert_called_with(['relation-ids', '--format=json',
                                         reltype])
        self.assertFalse(relation_type.called)

    @patch('subprocess.check_output')
    @patch('charmhelpers.core.hookenv.relation_id')
    def test_gets_related_units(self, relation_id, check_output):
        relid = 123
        units = ['foo', 'bar']
        relation_id.return_value = relid
        check_output.return_value = json.dumps(units).encode('UTF-8')

        result = hookenv.related_units()

        self.assertEqual(result, units)
        check_output.assert_called_with(['relation-list', '--format=json',
                                         '-r', relid])

    @patch('subprocess.check_output')
    @patch('charmhelpers.core.hookenv.relation_id')
    def test_gets_related_units_empty_array(self, relation_id, check_output):
        relid = str(123)
        units = []
        relation_id.return_value = relid
        check_output.return_value = json.dumps(None).encode('UTF-8')

        result = hookenv.related_units()

        self.assertEqual(result, units)
        check_output.assert_called_with(['relation-list', '--format=json',
                                         '-r', relid])

    @patch('subprocess.check_output')
    @patch('charmhelpers.core.hookenv.relation_id')
    def test_related_units_no_relation(self, relation_id, check_output):
        units = ['foo', 'bar']
        relation_id.return_value = None
        check_output.return_value = json.dumps(units).encode('UTF-8')

        result = hookenv.related_units()

        self.assertEqual(result, units)
        check_output.assert_called_with(['relation-list', '--format=json'])

    @patch('subprocess.check_output')
    @patch('charmhelpers.core.hookenv.relation_id')
    def test_gets_related_units_for_id(self, relation_id, check_output):
        relid = 123
        units = ['foo', 'bar']
        check_output.return_value = json.dumps(units).encode('UTF-8')

        result = hookenv.related_units(relid)

        self.assertEqual(result, units)
        check_output.assert_called_with(['relation-list', '--format=json',
                                         '-r', relid])
        self.assertFalse(relation_id.called)

    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('charmhelpers.core.hookenv.goal_state')
    @patch('charmhelpers.core.hookenv.has_juju_version')
    def test_gets_expected_peer_units(self, has_juju_version, goal_state,
                                      local_unit):
        has_juju_version.return_value = True
        goal_state.return_value = {
            'units': {
                'keystone/0': {
                    'status': 'active',
                    'since': '2018-09-27 11:38:28Z',
                },
                'keystone/1': {
                    'status': 'active',
                    'since': '2018-09-27 11:39:23Z',
                },
            },
        }
        local_unit.return_value = 'keystone/0'

        result = hookenv.expected_peer_units()

        self.assertIsInstance(result, types.GeneratorType)
        self.assertEqual(sorted(result), ['keystone/1'])
        has_juju_version.assertCalledOnceWith("2.4.0")
        local_unit.assertCalledOnceWith()

    @patch('charmhelpers.core.hookenv.has_juju_version')
    def test_gets_expected_peer_units_wrong_version(self, has_juju_version):
        has_juju_version.return_value = False

        def x():
            # local helper function to make testtools.TestCase.assertRaises
            # work with generator
            list(hookenv.expected_peer_units())

        self.assertRaises(NotImplementedError, x)
        has_juju_version.assertCalledOnceWith("2.4.0")

    @patch('charmhelpers.core.hookenv.goal_state')
    @patch('charmhelpers.core.hookenv.relation_type')
    @patch('charmhelpers.core.hookenv.has_juju_version')
    def test_gets_expected_related_units(self, has_juju_version, relation_type,
                                         goal_state):
        has_juju_version.return_value = True
        relation_type.return_value = 'identity-service'
        goal_state.return_value = {
            'relations': {
                'identity-service': {
                    'glance': {
                        'status': 'joined',
                        'since': '2018-09-27 11:37:16Z'
                    },
                    'glance/0': {
                        'status': 'active',
                        'since': '2018-09-27 11:27:19Z'
                    },
                    'glance/1': {
                        'status': 'active',
                        'since': '2018-09-27 11:27:34Z'
                    },
                },
            },
        }

        result = hookenv.expected_related_units()

        self.assertIsInstance(result, types.GeneratorType)
        self.assertEqual(sorted(result), ['glance/0', 'glance/1'])

    @patch('charmhelpers.core.hookenv.goal_state')
    @patch('charmhelpers.core.hookenv.has_juju_version')
    def test_gets_expected_related_units_for_type(self, has_juju_version,
                                                  goal_state):
        has_juju_version.return_value = True
        goal_state.return_value = {
            'relations': {
                'identity-service': {
                    'glance': {
                        'status': 'joined',
                        'since': '2018-09-27 11:37:16Z'
                    },
                    'glance/0': {
                        'status': 'active',
                        'since': '2018-09-27 11:27:19Z'
                    },
                    'glance/1': {
                        'status': 'active',
                        'since': '2018-09-27 11:27:34Z'
                    },
                },
            },
        }

        result = hookenv.expected_related_units('identity-service')

        self.assertIsInstance(result, types.GeneratorType)
        self.assertEqual(sorted(result), ['glance/0', 'glance/1'])

    @patch('charmhelpers.core.hookenv.has_juju_version')
    def test_gets_expected_related_units_wrong_version(self, has_juju_version):
        has_juju_version.return_value = False

        def x():
            # local helper function to make testtools.TestCase.assertRaises
            # work with generator
            list(hookenv.expected_related_units())

        self.assertRaises(NotImplementedError, x)
        has_juju_version.assertCalledOnceWith("2.4.4")

    @patch('charmhelpers.core.hookenv.os')
    def test_gets_the_remote_unit(self, os_):
        os_.environ = {
            'JUJU_REMOTE_UNIT': 'foo',
        }

        self.assertEqual(hookenv.remote_unit(), 'foo')

    @patch('charmhelpers.core.hookenv.os')
    def test_no_remote_unit(self, os_):
        os_.environ = {}
        self.assertEqual(hookenv.remote_unit(), None)

    @patch('charmhelpers.core.hookenv.remote_unit')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_gets_relation_for_unit(self, relation_get, remote_unit):
        unit = 'foo-unit'
        raw_relation = {
            'foo': 'bar',
        }
        remote_unit.return_value = unit
        relation_get.return_value = raw_relation

        result = hookenv.relation_for_unit()

        self.assertEqual(result['__unit__'], unit)
        self.assertEqual(result['foo'], 'bar')
        relation_get.assert_called_with(unit=unit, rid=None)

    @patch('charmhelpers.core.hookenv.remote_unit')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_gets_relation_for_unit_with_list(self, relation_get, remote_unit):
        unit = 'foo-unit'
        raw_relation = {
            'foo-list': 'one two three',
        }
        remote_unit.return_value = unit
        relation_get.return_value = raw_relation

        result = hookenv.relation_for_unit()

        self.assertEqual(result['__unit__'], unit)
        self.assertEqual(result['foo-list'], ['one', 'two', 'three'])
        relation_get.assert_called_with(unit=unit, rid=None)

    @patch('charmhelpers.core.hookenv.remote_unit')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_gets_relation_for_specific_unit(self, relation_get, remote_unit):
        unit = 'foo-unit'
        raw_relation = {
            'foo': 'bar',
        }
        relation_get.return_value = raw_relation

        result = hookenv.relation_for_unit(unit)

        self.assertEqual(result['__unit__'], unit)
        self.assertEqual(result['foo'], 'bar')
        relation_get.assert_called_with(unit=unit, rid=None)
        self.assertFalse(remote_unit.called)

    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.relation_for_unit')
    def test_gets_relations_for_id(self, relation_for_unit, related_units,
                                   relation_ids):
        relid = 123
        units = ['foo', 'bar']
        unit_data = [
            {'foo-item': 'bar-item'},
            {'foo-item2': 'bar-item2'},
        ]
        relation_ids.return_value = relid
        related_units.return_value = units
        relation_for_unit.side_effect = unit_data

        result = hookenv.relations_for_id()

        self.assertEqual(result[0]['__relid__'], relid)
        self.assertEqual(result[0]['foo-item'], 'bar-item')
        self.assertEqual(result[1]['__relid__'], relid)
        self.assertEqual(result[1]['foo-item2'], 'bar-item2')
        related_units.assert_called_with(relid)
        self.assertEqual(relation_for_unit.mock_calls, [
            call('foo', relid),
            call('bar', relid),
        ])

    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.relation_for_unit')
    def test_gets_relations_for_specific_id(self, relation_for_unit,
                                            related_units, relation_ids):
        relid = 123
        units = ['foo', 'bar']
        unit_data = [
            {'foo-item': 'bar-item'},
            {'foo-item2': 'bar-item2'},
        ]
        related_units.return_value = units
        relation_for_unit.side_effect = unit_data

        result = hookenv.relations_for_id(relid)

        self.assertEqual(result[0]['__relid__'], relid)
        self.assertEqual(result[0]['foo-item'], 'bar-item')
        self.assertEqual(result[1]['__relid__'], relid)
        self.assertEqual(result[1]['foo-item2'], 'bar-item2')
        related_units.assert_called_with(relid)
        self.assertEqual(relation_for_unit.mock_calls, [
            call('foo', relid),
            call('bar', relid),
        ])
        self.assertFalse(relation_ids.called)

    @patch('charmhelpers.core.hookenv.in_relation_hook')
    @patch('charmhelpers.core.hookenv.relation_type')
    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.relations_for_id')
    def test_gets_relations_for_type(self, relations_for_id, relation_ids,
                                     relation_type, in_relation_hook):
        reltype = 'foo-type'
        relids = [123, 234]
        relations = [
            [
                {'foo': 'bar'},
                {'foo2': 'bar2'},
            ],
            [
                {'FOO': 'BAR'},
                {'FOO2': 'BAR2'},
            ],
        ]
        is_in_relation = True

        relation_type.return_value = reltype
        relation_ids.return_value = relids
        relations_for_id.side_effect = relations
        in_relation_hook.return_value = is_in_relation

        result = hookenv.relations_of_type()

        self.assertEqual(result[0]['__relid__'], 123)
        self.assertEqual(result[0]['foo'], 'bar')
        self.assertEqual(result[1]['__relid__'], 123)
        self.assertEqual(result[1]['foo2'], 'bar2')
        self.assertEqual(result[2]['__relid__'], 234)
        self.assertEqual(result[2]['FOO'], 'BAR')
        self.assertEqual(result[3]['__relid__'], 234)
        self.assertEqual(result[3]['FOO2'], 'BAR2')
        relation_ids.assert_called_with(reltype)
        self.assertEqual(relations_for_id.mock_calls, [
            call(123),
            call(234),
        ])

    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('charmhelpers.core.hookenv.relation_types')
    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_gets_relations(self, relation_get, related_units,
                            relation_ids, relation_types, local_unit):
        local_unit.return_value = 'u0'
        relation_types.return_value = ['t1', 't2']
        relation_ids.return_value = ['i1']
        related_units.return_value = ['u1', 'u2']
        relation_get.return_value = {'key': 'val'}

        result = hookenv.relations()

        self.assertEqual(result, {
            't1': {
                'i1': {
                    'u0': {'key': 'val'},
                    'u1': {'key': 'val'},
                    'u2': {'key': 'val'},
                },
            },
            't2': {
                'i1': {
                    'u0': {'key': 'val'},
                    'u1': {'key': 'val'},
                    'u2': {'key': 'val'},
                },
            },
        })

    @patch('charmhelpers.core.hookenv.relation_set')
    @patch('charmhelpers.core.hookenv.relation_get')
    @patch('charmhelpers.core.hookenv.local_unit')
    def test_relation_clear(self, local_unit,
                            relation_get,
                            relation_set):
        local_unit.return_value = 'local-unit'
        relation_get.return_value = {
            'private-address': '10.5.0.1',
            'foo': 'bar',
            'public-address': '146.192.45.6'
        }
        hookenv.relation_clear('relation:1')
        relation_get.assert_called_with(rid='relation:1',
                                        unit='local-unit')
        relation_set.assert_called_with(
            relation_id='relation:1',
            **{'private-address': '10.5.0.1',
               'foo': None,
               'public-address': '146.192.45.6'})

    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_is_relation_made(self, relation_get, related_units,
                              relation_ids):
        relation_get.return_value = 'hostname'
        related_units.return_value = ['test/1']
        relation_ids.return_value = ['test:0']
        self.assertTrue(hookenv.is_relation_made('test'))
        relation_get.assert_called_with('private-address',
                                        rid='test:0', unit='test/1')

    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_is_relation_made_multi_unit(self, relation_get, related_units,
                                         relation_ids):
        relation_get.side_effect = [None, 'hostname']
        related_units.return_value = ['test/1', 'test/2']
        relation_ids.return_value = ['test:0']
        self.assertTrue(hookenv.is_relation_made('test'))

    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_is_relation_made_different_key(self,
                                            relation_get, related_units,
                                            relation_ids):
        relation_get.return_value = 'hostname'
        related_units.return_value = ['test/1']
        relation_ids.return_value = ['test:0']
        self.assertTrue(hookenv.is_relation_made('test', keys='auth'))
        relation_get.assert_called_with('auth',
                                        rid='test:0', unit='test/1')

    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_is_relation_made_multiple_keys(self,
                                            relation_get, related_units,
                                            relation_ids):
        relation_get.side_effect = ['password', 'hostname']
        related_units.return_value = ['test/1']
        relation_ids.return_value = ['test:0']
        self.assertTrue(hookenv.is_relation_made('test',
                                                 keys=['auth', 'host']))
        relation_get.assert_has_calls(
            [call('auth', rid='test:0', unit='test/1'),
             call('host', rid='test:0', unit='test/1')]
        )

    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_is_relation_made_not_made(self,
                                       relation_get, related_units,
                                       relation_ids):
        relation_get.return_value = None
        related_units.return_value = ['test/1']
        relation_ids.return_value = ['test:0']
        self.assertFalse(hookenv.is_relation_made('test'))

    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.relation_get')
    def test_is_relation_made_not_made_multiple_keys(self,
                                                     relation_get,
                                                     related_units,
                                                     relation_ids):
        relation_get.side_effect = ['password', None]
        related_units.return_value = ['test/1']
        relation_ids.return_value = ['test:0']
        self.assertFalse(hookenv.is_relation_made('test',
                                                  keys=['auth', 'host']))
        relation_get.assert_has_calls(
            [call('auth', rid='test:0', unit='test/1'),
             call('host', rid='test:0', unit='test/1')]
        )

    @patch('charmhelpers.core.hookenv.config')
    @patch('charmhelpers.core.hookenv.relation_type')
    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('charmhelpers.core.hookenv.relation_id')
    @patch('charmhelpers.core.hookenv.relations')
    @patch('charmhelpers.core.hookenv.relation_get')
    @patch('charmhelpers.core.hookenv.os')
    def test_gets_execution_environment(self, os_, relations_get,
                                        relations, relation_id, local_unit,
                                        relation_type, config):
        config.return_value = 'some-config'
        relation_type.return_value = 'some-type'
        local_unit.return_value = 'some-unit'
        relation_id.return_value = 'some-id'
        relations.return_value = 'all-relations'
        relations_get.return_value = 'some-relations'
        os_.environ = 'some-environment'

        result = hookenv.execution_environment()

        self.assertEqual(result, {
            'conf': 'some-config',
            'reltype': 'some-type',
            'unit': 'some-unit',
            'relid': 'some-id',
            'rel': 'some-relations',
            'rels': 'all-relations',
            'env': 'some-environment',
        })

    @patch('charmhelpers.core.hookenv.config')
    @patch('charmhelpers.core.hookenv.relation_type')
    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('charmhelpers.core.hookenv.relation_id')
    @patch('charmhelpers.core.hookenv.relations')
    @patch('charmhelpers.core.hookenv.relation_get')
    @patch('charmhelpers.core.hookenv.os')
    def test_gets_execution_environment_no_relation(
            self, os_, relations_get, relations, relation_id,
            local_unit, relation_type, config):
        config.return_value = 'some-config'
        relation_type.return_value = 'some-type'
        local_unit.return_value = 'some-unit'
        relation_id.return_value = None
        relations.return_value = 'all-relations'
        relations_get.return_value = 'some-relations'
        os_.environ = 'some-environment'

        result = hookenv.execution_environment()

        self.assertEqual(result, {
            'conf': 'some-config',
            'unit': 'some-unit',
            'rels': 'all-relations',
            'env': 'some-environment',
        })

    @patch('charmhelpers.core.hookenv.remote_service_name')
    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.os')
    def test_gets_the_relation_id(self, os_, relation_ids, remote_service_name):
        os_.environ = {
            'JUJU_RELATION_ID': 'foo',
        }

        self.assertEqual(hookenv.relation_id(), 'foo')

        relation_ids.return_value = ['r:1', 'r:2']
        remote_service_name.side_effect = ['other', 'service']
        self.assertEqual(hookenv.relation_id('rel', 'service/0'), 'r:2')
        relation_ids.assert_called_once_with('rel')
        self.assertEqual(remote_service_name.call_args_list, [
            call('r:1'),
            call('r:2'),
        ])
        remote_service_name.side_effect = ['other', 'service']
        self.assertEqual(hookenv.relation_id('rel', 'service'), 'r:2')

    @patch('charmhelpers.core.hookenv.os')
    def test_relation_id_none_if_no_env(self, os_):
        os_.environ = {}
        self.assertEqual(hookenv.relation_id(), None)

    @patch('subprocess.check_output')
    def test_gets_relation(self, check_output):
        data = {"foo": "BAR"}
        check_output.return_value = json.dumps(data).encode('UTF-8')
        result = hookenv.relation_get()

        self.assertEqual(result['foo'], 'BAR')
        check_output.assert_called_with(['relation-get', '--format=json', '-'])

    @patch('charmhelpers.core.hookenv.subprocess')
    def test_relation_get_none(self, mock_subprocess):
        mock_subprocess.check_output.return_value = b'null'

        result = hookenv.relation_get()

        self.assertIsNone(result)

    @patch('charmhelpers.core.hookenv.subprocess')
    def test_relation_get_calledprocesserror(self, mock_subprocess):
        """relation-get called outside a relation will errors without id."""
        mock_subprocess.check_output.side_effect = CalledProcessError(
            2, '/foo/bin/relation-get'
            'no relation id specified')

        result = hookenv.relation_get()

        self.assertIsNone(result)

    @patch('charmhelpers.core.hookenv.subprocess')
    def test_relation_get_calledprocesserror_other(self, mock_subprocess):
        """relation-get can fail for other more serious errors."""
        mock_subprocess.check_output.side_effect = CalledProcessError(
            1, '/foo/bin/relation-get'
            'connection refused')

        self.assertRaises(CalledProcessError, hookenv.relation_get)

    @patch('subprocess.check_output')
    def test_gets_relation_with_scope(self, check_output):
        check_output.return_value = json.dumps('bar').encode('UTF-8')

        result = hookenv.relation_get(attribute='baz-scope')

        self.assertEqual(result, 'bar')
        check_output.assert_called_with(['relation-get', '--format=json',
                                         'baz-scope'])

    @patch('subprocess.check_output')
    def test_gets_missing_relation_with_scope(self, check_output):
        check_output.return_value = b""

        result = hookenv.relation_get(attribute='baz-scope')

        self.assertEqual(result, None)
        check_output.assert_called_with(['relation-get', '--format=json',
                                         'baz-scope'])

    @patch('subprocess.check_output')
    def test_gets_relation_with_unit_name(self, check_output):
        check_output.return_value = json.dumps('BAR').encode('UTF-8')

        result = hookenv.relation_get(attribute='baz-scope', unit='baz-unit')

        self.assertEqual(result, 'BAR')
        check_output.assert_called_with(['relation-get', '--format=json',
                                         'baz-scope', 'baz-unit'])

    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('subprocess.check_call')
    @patch('subprocess.check_output')
    def test_relation_set_flushes_local_unit_cache(self, check_output,
                                                   check_call, local_unit):
        check_output.return_value = json.dumps('BAR').encode('UTF-8')
        local_unit.return_value = 'baz_unit'
        hookenv.relation_get(attribute='baz_scope', unit='baz_unit')
        hookenv.relation_get(attribute='bar_scope')
        self.assertTrue(len(hookenv.cache) == 2)
        check_output.return_value = ""
        hookenv.relation_set(baz_scope='hello')
        # relation_set should flush any entries for local_unit
        self.assertTrue(len(hookenv.cache) == 1)

    @patch('subprocess.check_output')
    def test_gets_relation_with_relation_id(self, check_output):
        check_output.return_value = json.dumps('BAR').encode('UTF-8')

        result = hookenv.relation_get(attribute='baz-scope', unit='baz-unit',
                                      rid=123)

        self.assertEqual(result, 'BAR')
        check_output.assert_called_with(['relation-get', '--format=json', '-r',
                                         123, 'baz-scope', 'baz-unit'])

    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('subprocess.check_output')
    @patch('subprocess.check_call')
    def test_sets_relation_with_kwargs(self, check_call_, check_output,
                                       local_unit):
        hookenv.relation_set(foo="bar")
        check_call_.assert_called_with(['relation-set', 'foo=bar'])

    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('subprocess.check_output')
    @patch('subprocess.check_call')
    def test_sets_relation_with_dict(self, check_call_, check_output,
                                     local_unit):
        hookenv.relation_set(relation_settings={"foo": "bar"})
        check_call_.assert_called_with(['relation-set', 'foo=bar'])

    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('subprocess.check_output')
    @patch('subprocess.check_call')
    def test_sets_relation_with_relation_id(self, check_call_, check_output,
                                            local_unit):
        hookenv.relation_set(relation_id="foo", bar="baz")
        check_call_.assert_called_with(['relation-set', '-r', 'foo',
                                        'bar=baz'])

    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('subprocess.check_output')
    @patch('subprocess.check_call')
    def test_sets_relation_with_missing_value(self, check_call_, check_output,
                                              local_unit):
        hookenv.relation_set(foo=None)
        check_call_.assert_called_with(['relation-set', 'foo='])

    @patch('charmhelpers.core.hookenv.local_unit', MagicMock())
    @patch('os.remove')
    @patch('subprocess.check_output')
    @patch('subprocess.check_call')
    def test_relation_set_file(self, check_call, check_output, remove):
        """If relation-set accepts a --file parameter, it's used.

        Juju 1.23.2 introduced a --file parameter, which means you can
        pass the data through a file. Not using --file would make
        relation_set break if the relation data is too big.
        """
        # check_output(["relation-set", "--help"]) is used to determine
        # whether we can pass --file to it.
        check_output.return_value = "--file"
        hookenv.relation_set(foo="bar")
        check_output.assert_called_with(
            ["relation-set", "--help"], universal_newlines=True)
        # relation-set is called with relation-set --file <temp_file>
        # with data as YAML and the temp_file is then removed.
        self.assertEqual(1, len(check_call.call_args[0]))
        command = check_call.call_args[0][0]
        self.assertEqual(3, len(command))
        self.assertEqual("relation-set", command[0])
        self.assertEqual("--file", command[1])
        temp_file = command[2]
        with open(temp_file, "r") as f:
            self.assertEqual("{foo: bar}", f.read().strip())
        remove.assert_called_with(temp_file)

    @patch('charmhelpers.core.hookenv.local_unit', MagicMock())
    @patch('os.remove')
    @patch('subprocess.check_output')
    @patch('subprocess.check_call')
    def test_relation_set_file_non_str(self, check_call, check_output, remove):
        """If relation-set accepts a --file parameter, it's used.

        Any value that is not a string is converted to a string before encoding
        the settings to YAML.
        """
        # check_output(["relation-set", "--help"]) is used to determine
        # whether we can pass --file to it.
        check_output.return_value = "--file"
        hookenv.relation_set(foo={"bar": 1})
        check_output.assert_called_with(
            ["relation-set", "--help"], universal_newlines=True)
        # relation-set is called with relation-set --file <temp_file>
        # with data as YAML and the temp_file is then removed.
        self.assertEqual(1, len(check_call.call_args[0]))
        command = check_call.call_args[0][0]
        self.assertEqual(3, len(command))
        self.assertEqual("relation-set", command[0])
        self.assertEqual("--file", command[1])
        temp_file = command[2]
        with open(temp_file, "r") as f:
            self.assertEqual("{foo: '{''bar'': 1}'}", f.read().strip())
        remove.assert_called_with(temp_file)

    def test_lists_relation_types(self):
        open_ = mock_open()
        open_.return_value = io.BytesIO(CHARM_METADATA)

        with patch('charmhelpers.core.hookenv.open', open_, create=True):
            with patch.dict('os.environ', {'CHARM_DIR': '/var/empty'}):
                reltypes = set(hookenv.relation_types())
        open_.assert_called_once_with('/var/empty/metadata.yaml')
        self.assertEqual(set(('testreqs', 'testprov', 'testpeer')), reltypes)

    def test_metadata(self):
        open_ = mock_open()
        open_.return_value = io.BytesIO(CHARM_METADATA)

        with patch('charmhelpers.core.hookenv.open', open_, create=True):
            with patch.dict('os.environ', {'CHARM_DIR': '/var/empty'}):
                metadata = hookenv.metadata()
        self.assertEqual(metadata, yaml.safe_load(CHARM_METADATA))

    @patch('charmhelpers.core.hookenv.relation_ids')
    @patch('charmhelpers.core.hookenv.metadata')
    def test_peer_relation_id(self, metadata, relation_ids):
        metadata.return_value = {'peers': {sentinel.peer_relname: {}}}
        relation_ids.return_value = [sentinel.pid1, sentinel.pid2]
        self.assertEqual(hookenv.peer_relation_id(), sentinel.pid1)
        relation_ids.assert_called_once_with(sentinel.peer_relname)

    def test_charm_name(self):
        open_ = mock_open()
        open_.return_value = io.BytesIO(CHARM_METADATA)

        with patch('charmhelpers.core.hookenv.open', open_, create=True):
            with patch.dict('os.environ', {'CHARM_DIR': '/var/empty'}):
                charm_name = hookenv.charm_name()
        self.assertEqual("testmock", charm_name)

    @patch('subprocess.check_call')
    def test_opens_port(self, check_call_):
        hookenv.open_port(443, "TCP")
        hookenv.open_port(80)
        hookenv.open_port(100, "UDP")
        hookenv.open_port(0, "ICMP")
        calls = [
            call(['open-port', '443/TCP']),
            call(['open-port', '80/TCP']),
            call(['open-port', '100/UDP']),
            call(['open-port', 'ICMP']),
        ]
        check_call_.assert_has_calls(calls)

    @patch('subprocess.check_call')
    def test_closes_port(self, check_call_):
        hookenv.close_port(443, "TCP")
        hookenv.close_port(80)
        hookenv.close_port(100, "UDP")
        hookenv.close_port(0, "ICMP")
        calls = [
            call(['close-port', '443/TCP']),
            call(['close-port', '80/TCP']),
            call(['close-port', '100/UDP']),
            call(['close-port', 'ICMP']),
        ]
        check_call_.assert_has_calls(calls)

    @patch('subprocess.check_call')
    def test_opens_ports(self, check_call_):
        hookenv.open_ports(443, 447, "TCP")
        hookenv.open_ports(80, 91)
        hookenv.open_ports(100, 200, "UDP")
        calls = [
            call(['open-port', '443-447/TCP']),
            call(['open-port', '80-91/TCP']),
            call(['open-port', '100-200/UDP']),
        ]
        check_call_.assert_has_calls(calls)

    @patch('subprocess.check_call')
    def test_closes_ports(self, check_call_):
        hookenv.close_ports(443, 447, "TCP")
        hookenv.close_ports(80, 91)
        hookenv.close_ports(100, 200, "UDP")
        calls = [
            call(['close-port', '443-447/TCP']),
            call(['close-port', '80-91/TCP']),
            call(['close-port', '100-200/UDP']),
        ]
        check_call_.assert_has_calls(calls)

    @patch('subprocess.check_output')
    def test_gets_opened_ports(self, check_output_):
        prts = ['8080/tcp', '8081-8083/tcp']
        check_output_.return_value = json.dumps(prts).encode('UTF-8')
        self.assertEqual(hookenv.opened_ports(), prts)
        check_output_.assert_called_with(['opened-ports', '--format=json'])

    @patch('subprocess.check_output')
    def test_gets_unit_attribute(self, check_output_):
        check_output_.return_value = json.dumps('bar').encode('UTF-8')
        self.assertEqual(hookenv.unit_get('foo'), 'bar')
        check_output_.assert_called_with(['unit-get', '--format=json', 'foo'])

    @patch('subprocess.check_output')
    def test_gets_missing_unit_attribute(self, check_output_):
        check_output_.return_value = b""
        self.assertEqual(hookenv.unit_get('foo'), None)
        check_output_.assert_called_with(['unit-get', '--format=json', 'foo'])

    def test_cached_decorator(self):
        class Unserializable(object):
            def __str__(self):
                return 'unserializable'

        unserializable = Unserializable()
        calls = []
        values = {
            'hello': 'world',
            'foo': 'bar',
            'baz': None,
            unserializable: 'qux',
        }

        @hookenv.cached
        def cache_function(attribute):
            calls.append(attribute)
            return values[attribute]

        self.assertEquals(cache_function('hello'), 'world')
        self.assertEquals(cache_function('hello'), 'world')
        self.assertEquals(cache_function('foo'), 'bar')
        self.assertEquals(cache_function('baz'), None)
        self.assertEquals(cache_function('baz'), None)
        self.assertEquals(cache_function(unserializable), 'qux')
        self.assertEquals(calls, ['hello', 'foo', 'baz', unserializable])

    def test_gets_charm_dir(self):
        with patch.dict('os.environ', {}):
            self.assertEqual(hookenv.charm_dir(), None)
        with patch.dict('os.environ', {'CHARM_DIR': '/var/empty'}):
            self.assertEqual(hookenv.charm_dir(), '/var/empty')
        with patch.dict('os.environ', {'JUJU_CHARM_DIR': '/var/another'}):
            self.assertEqual(hookenv.charm_dir(), '/var/another')

    @patch('subprocess.check_output')
    def test_resource_get_unsupported(self, check_output_):
        check_output_.side_effect = OSError(2, 'resource-get')
        self.assertRaises(NotImplementedError, hookenv.resource_get, 'foo')

    @patch('subprocess.check_output')
    def test_resource_get(self, check_output_):
        check_output_.return_value = b'/tmp/file'
        self.assertEqual(hookenv.resource_get('file'), '/tmp/file')
        check_output_.side_effect = CalledProcessError(
            1, '/foo/bin/resource-get',
            'error: could not download resource: resource#foo/file not found')
        self.assertFalse(hookenv.resource_get('no-file'))
        self.assertFalse(hookenv.resource_get(None))

    @patch('subprocess.check_output')
    def test_goal_state_unsupported(self, check_output_):
        check_output_.side_effect = OSError(2, 'goal-state')
        self.assertRaises(NotImplementedError, hookenv.goal_state)

    @patch('subprocess.check_output')
    def test_goal_state(self, check_output_):
        expect = {
            'units': {},
            'relations': {},
        }
        check_output_.return_value = json.dumps(expect).encode('UTF-8')
        result = hookenv.goal_state()

        self.assertEqual(result, expect)
        check_output_.assert_called_with(['goal-state', '--format=json'])

    @patch('subprocess.check_output')
    def test_is_leader_unsupported(self, check_output_):
        check_output_.side_effect = OSError(2, 'is-leader')
        self.assertRaises(NotImplementedError, hookenv.is_leader)

    @patch('subprocess.check_output')
    def test_is_leader(self, check_output_):
        check_output_.return_value = b'false'
        self.assertFalse(hookenv.is_leader())
        check_output_.return_value = b'true'
        self.assertTrue(hookenv.is_leader())

    @patch('subprocess.check_call')
    def test_payload_register(self, check_call_):
        hookenv.payload_register('monitoring', 'docker', 'abc123')
        check_call_.assert_called_with(['payload-register', 'monitoring',
                                        'docker', 'abc123'])

    @patch('subprocess.check_call')
    def test_payload_unregister(self, check_call_):
        hookenv.payload_unregister('monitoring', 'abc123')
        check_call_.assert_called_with(['payload-unregister', 'monitoring',
                                        'abc123'])

    @patch('subprocess.check_call')
    def test_payload_status_set(self, check_call_):
        hookenv.payload_status_set('monitoring', 'abc123', 'Running')
        check_call_.assert_called_with(['payload-status-set', 'monitoring',
                                        'abc123', 'Running'])

    @patch('subprocess.check_call')
    def test_application_version_set(self, check_call_):
        hookenv.application_version_set('v1.2.3')
        check_call_.assert_called_with(['application-version-set', 'v1.2.3'])

    @patch.object(os, 'getenv')
    @patch.object(hookenv, 'log')
    def test_env_proxy_settings_juju_charm_all_selected(self, faux_log,
                                                        get_env):
        expected_settings = {
            'HTTP_PROXY': 'http://squid.internal:3128',
            'http_proxy': 'http://squid.internal:3128',
            'HTTPS_PROXY': 'https://squid.internals:3128',
            'https_proxy': 'https://squid.internals:3128',
            'NO_PROXY': '192.0.2.0/24,198.51.100.0/24,.bar.com',
            'no_proxy': '192.0.2.0/24,198.51.100.0/24,.bar.com',
            'FTP_PROXY': 'ftp://ftp.internal:21',
            'ftp_proxy': 'ftp://ftp.internal:21',
        }

        def get_env_side_effect(var):
            return {
                'HTTP_PROXY': None,
                'HTTPS_PROXY': None,
                'NO_PROXY': None,
                'FTP_PROXY': None,
                'JUJU_CHARM_HTTP_PROXY': 'http://squid.internal:3128',
                'JUJU_CHARM_HTTPS_PROXY': 'https://squid.internals:3128',
                'JUJU_CHARM_FTP_PROXY': 'ftp://ftp.internal:21',
                'JUJU_CHARM_NO_PROXY': '192.0.2.0/24,198.51.100.0/24,.bar.com'
            }[var]
        get_env.side_effect = get_env_side_effect

        proxy_settings = hookenv.env_proxy_settings()
        get_env.assert_has_calls([call("HTTP_PROXY"),
                                 call("HTTPS_PROXY"),
                                 call("NO_PROXY"),
                                 call("FTP_PROXY"),
                                 call("JUJU_CHARM_HTTP_PROXY"),
                                 call("JUJU_CHARM_HTTPS_PROXY"),
                                 call("JUJU_CHARM_FTP_PROXY"),
                                 call("JUJU_CHARM_NO_PROXY")],
                                 any_order=True)
        self.assertEqual(expected_settings, proxy_settings)
        # Verify that we logged a warning about the cidr in NO_PROXY.
        faux_log.assert_called_with(hookenv.RANGE_WARNING,
                                    level=hookenv.WARNING)

    @patch.object(os, 'getenv')
    def test_env_proxy_settings_legacy_https(self, get_env):
        expected_settings = {
            'HTTPS_PROXY': 'http://squid.internal:3128',
            'https_proxy': 'http://squid.internal:3128',
        }

        def get_env_side_effect(var):
            return {
                'HTTPS_PROXY': 'http://squid.internal:3128',
                'JUJU_CHARM_HTTPS_PROXY': None,
            }[var]
        get_env.side_effect = get_env_side_effect

        proxy_settings = hookenv.env_proxy_settings(['https'])
        get_env.assert_has_calls([call("HTTPS_PROXY"),
                                 call("JUJU_CHARM_HTTPS_PROXY")],
                                 any_order=True)
        self.assertEqual(expected_settings, proxy_settings)

    @patch.object(os, 'getenv')
    def test_env_proxy_settings_juju_charm_https(self, get_env):
        expected_settings = {
            'HTTPS_PROXY': 'http://squid.internal:3128',
            'https_proxy': 'http://squid.internal:3128',
        }

        def get_env_side_effect(var):
            return {
                'HTTPS_PROXY': None,
                'JUJU_CHARM_HTTPS_PROXY': 'http://squid.internal:3128',
            }[var]
        get_env.side_effect = get_env_side_effect

        proxy_settings = hookenv.env_proxy_settings(['https'])
        get_env.assert_has_calls([call("HTTPS_PROXY"),
                                 call("JUJU_CHARM_HTTPS_PROXY")],
                                 any_order=True)
        self.assertEqual(expected_settings, proxy_settings)

    @patch.object(os, 'getenv')
    def test_env_proxy_settings_legacy_http(self, get_env):
        expected_settings = {
            'HTTP_PROXY': 'http://squid.internal:3128',
            'http_proxy': 'http://squid.internal:3128',
        }

        def get_env_side_effect(var):
            return {
                'HTTP_PROXY': 'http://squid.internal:3128',
                'JUJU_CHARM_HTTP_PROXY': None,
            }[var]
        get_env.side_effect = get_env_side_effect

        proxy_settings = hookenv.env_proxy_settings(['http'])
        get_env.assert_has_calls([call("HTTP_PROXY"),
                                 call("JUJU_CHARM_HTTP_PROXY")],
                                 any_order=True)
        self.assertEqual(expected_settings, proxy_settings)

    @patch.object(os, 'getenv')
    def test_env_proxy_settings_juju_charm_http(self, get_env):
        expected_settings = {
            'HTTP_PROXY': 'http://squid.internal:3128',
            'http_proxy': 'http://squid.internal:3128',
        }

        def get_env_side_effect(var):
            return {
                'HTTP_PROXY': None,
                'JUJU_CHARM_HTTP_PROXY': 'http://squid.internal:3128',
            }[var]
        get_env.side_effect = get_env_side_effect

        proxy_settings = hookenv.env_proxy_settings(['http'])
        get_env.assert_has_calls([call("HTTP_PROXY"),
                                 call("JUJU_CHARM_HTTP_PROXY")],
                                 any_order=True)
        self.assertEqual(expected_settings, proxy_settings)


class HooksTest(TestCase):
    def setUp(self):
        super(HooksTest, self).setUp()

        _clean_globals()
        self.addCleanup(_clean_globals)

        charm_dir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(charm_dir))
        patcher = patch.object(hookenv, 'charm_dir', lambda: charm_dir)
        self.addCleanup(patcher.stop)
        patcher.start()

        config = hookenv.Config({})

        def _mock_config(scope=None):
            return config if scope is None else config[scope]
        patcher = patch.object(hookenv, 'config', _mock_config)
        self.addCleanup(patcher.stop)
        patcher.start()

    def test_config_saved_after_execute(self):
        config = hookenv.config()
        config.implicit_save = True

        foo = MagicMock()
        hooks = hookenv.Hooks()
        hooks.register('foo', foo)
        hooks.execute(['foo', 'some', 'other', 'args'])
        self.assertTrue(os.path.exists(config.path))

    def test_config_not_saved_after_execute(self):
        config = hookenv.config()
        config.implicit_save = False

        foo = MagicMock()
        hooks = hookenv.Hooks()
        hooks.register('foo', foo)
        hooks.execute(['foo', 'some', 'other', 'args'])
        self.assertFalse(os.path.exists(config.path))

    def test_config_save_disabled(self):
        config = hookenv.config()
        config.implicit_save = True

        foo = MagicMock()
        hooks = hookenv.Hooks(config_save=False)
        hooks.register('foo', foo)
        hooks.execute(['foo', 'some', 'other', 'args'])
        self.assertFalse(os.path.exists(config.path))

    def test_runs_a_registered_function(self):
        foo = MagicMock()
        hooks = hookenv.Hooks()
        hooks.register('foo', foo)

        hooks.execute(['foo', 'some', 'other', 'args'])

        foo.assert_called_with()

    def test_cannot_run_unregistered_function(self):
        foo = MagicMock()
        hooks = hookenv.Hooks()
        hooks.register('foo', foo)

        self.assertRaises(hookenv.UnregisteredHookError, hooks.execute,
                          ['bar'])

    def test_can_run_a_decorated_function_as_one_or_more_hooks(self):
        execs = []
        hooks = hookenv.Hooks()

        @hooks.hook('bar', 'baz')
        def func():
            execs.append(True)

        hooks.execute(['bar'])
        hooks.execute(['baz'])
        self.assertRaises(hookenv.UnregisteredHookError, hooks.execute,
                          ['brew'])
        self.assertEqual(execs, [True, True])

    def test_can_run_a_decorated_function_as_itself(self):
        execs = []
        hooks = hookenv.Hooks()

        @hooks.hook()
        def func():
            execs.append(True)

        hooks.execute(['func'])
        self.assertRaises(hookenv.UnregisteredHookError, hooks.execute,
                          ['brew'])
        self.assertEqual(execs, [True])

    def test_magic_underscores(self):
        # Juju hook names use hypens as separators. Python functions use
        # underscores. If explicit names have not been provided, hooks
        # are registered with both the function name and the function
        # name with underscores replaced with hypens for convenience.
        execs = []
        hooks = hookenv.Hooks()

        @hooks.hook()
        def call_me_maybe():
            execs.append(True)

        hooks.execute(['call-me-maybe'])
        hooks.execute(['call_me_maybe'])
        self.assertEqual(execs, [True, True])

    @patch('charmhelpers.core.hookenv.local_unit')
    def test_gets_service_name(self, _unit):
        _unit.return_value = 'mysql/3'
        self.assertEqual(hookenv.service_name(), 'mysql')

    @patch('charmhelpers.core.hookenv.related_units')
    @patch('charmhelpers.core.hookenv.remote_unit')
    def test_gets_remote_service_name(self, remote_unit, related_units):
        remote_unit.return_value = 'mysql/3'
        related_units.return_value = ['pgsql/0', 'pgsql/1']
        self.assertEqual(hookenv.remote_service_name(), 'mysql')
        self.assertEqual(hookenv.remote_service_name('pgsql:1'), 'pgsql')

    def test_gets_hook_name(self):
        with patch.dict(os.environ, JUJU_HOOK_NAME='hook'):
            self.assertEqual(hookenv.hook_name(), 'hook')
        with patch('sys.argv', ['other-hook']):
            self.assertEqual(hookenv.hook_name(), 'other-hook')

    @patch('subprocess.check_output')
    def test_action_get_with_key(self, check_output):
        action_data = 'bar'
        check_output.return_value = json.dumps(action_data).encode('UTF-8')

        result = hookenv.action_get(key='foo')

        self.assertEqual(result, 'bar')
        check_output.assert_called_with(['action-get', 'foo', '--format=json'])

    @patch('subprocess.check_output')
    def test_action_get_without_key(self, check_output):
        check_output.return_value = json.dumps(dict(foo='bar')).encode('UTF-8')

        result = hookenv.action_get()

        self.assertEqual(result['foo'], 'bar')
        check_output.assert_called_with(['action-get', '--format=json'])

    @patch('subprocess.check_call')
    def test_action_set(self, check_call):
        values = {'foo': 'bar', 'fooz': 'barz'}
        hookenv.action_set(values)
        # The order of the key/value pairs can change, so sort them before test
        called_args = check_call.call_args_list[0][0][0]
        called_args.pop(0)
        called_args.sort()
        self.assertEqual(called_args, ['foo=bar', 'fooz=barz'])

    @patch('subprocess.check_call')
    def test_action_fail(self, check_call):
        message = "Ooops, the action failed"
        hookenv.action_fail(message)
        check_call.assert_called_with(['action-fail', message])

    def test_status_set_invalid_state(self):
        self.assertRaises(ValueError, hookenv.status_set, 'random', 'message')

    @patch('subprocess.call')
    def test_status(self, call):
        call.return_value = 0
        hookenv.status_set('active', 'Everything is Awesome!')
        call.assert_called_with(['status-set', 'active', 'Everything is Awesome!'])

    @patch('subprocess.call')
    @patch.object(hookenv, 'log')
    def test_status_enoent(self, log, call):
        call.side_effect = OSError(2, 'fail')
        hookenv.status_set('active', 'Everything is Awesome!')
        log.assert_called_with('status-set failed: active Everything is Awesome!', level='INFO')

    @patch('subprocess.call')
    @patch.object(hookenv, 'log')
    def test_status_statuscmd_fail(self, log, call):
        call.side_effect = OSError(3, 'fail')
        self.assertRaises(OSError, hookenv.status_set, 'active', 'msg')
        call.assert_called_with(['status-set', 'active', 'msg'])

    @patch('subprocess.check_output')
    def test_status_get(self, check_output):
        check_output.return_value = json.dumps(
            {"message": "foo",
             "status": "active",
             "status-data": {}}).encode("UTF-8")
        result = hookenv.status_get()
        self.assertEqual(result, ('active', 'foo'))
        check_output.assert_called_with(
            ['status-get', "--format=json", "--include-data"])

    @patch('subprocess.check_output')
    def test_status_get_nostatus(self, check_output):
        check_output.side_effect = OSError(2, 'fail')
        result = hookenv.status_get()
        self.assertEqual(result, ('unknown', ''))

    @patch('subprocess.check_output')
    def test_status_get_status_error(self, check_output):
        check_output.side_effect = OSError(3, 'fail')
        self.assertRaises(OSError, hookenv.status_get)

    @patch('subprocess.check_output')
    @patch('glob.glob')
    def test_juju_version(self, glob, check_output):
        glob.return_value = [sentinel.jujud]
        check_output.return_value = '1.23.3.1-trusty-amd64\n'
        self.assertEqual(hookenv.juju_version(), '1.23.3.1-trusty-amd64')
        # Per https://bugs.launchpad.net/juju-core/+bug/1455368/comments/1
        glob.assert_called_once_with('/var/lib/juju/tools/machine-*/jujud')
        check_output.assert_called_once_with([sentinel.jujud, 'version'],
                                             universal_newlines=True)

    @patch('charmhelpers.core.hookenv.juju_version')
    def test_has_juju_version(self, juju_version):
        juju_version.return_value = '1.23.1.2.3.4.5-with-a-cherry-on-top.amd64'
        self.assertTrue(hookenv.has_juju_version('1.23'))
        self.assertTrue(hookenv.has_juju_version('1.23.1'))
        self.assertTrue(hookenv.has_juju_version('1.23.1.1'))
        self.assertFalse(hookenv.has_juju_version('1.23.2.1'))
        self.assertFalse(hookenv.has_juju_version('1.24'))

        juju_version.return_value = '1.24-beta5.1-trusty-amd64'
        self.assertTrue(hookenv.has_juju_version('1.23'))
        self.assertTrue(hookenv.has_juju_version('1.24'))  # Better if this was false!
        self.assertTrue(hookenv.has_juju_version('1.24-beta5'))
        self.assertTrue(hookenv.has_juju_version('1.24-beta5.1'))
        self.assertFalse(hookenv.has_juju_version('1.25'))
        self.assertTrue(hookenv.has_juju_version('1.18-backport6'))

    @patch.object(hookenv, 'relation_to_role_and_interface')
    def test_relation_to_interface(self, rtri):
        rtri.return_value = (None, 'foo')
        self.assertEqual(hookenv.relation_to_interface('rel'), 'foo')

    @patch.object(hookenv, 'metadata')
    def test_relation_to_role_and_interface(self, metadata):
        metadata.return_value = {
            'provides': {
                'pro-rel': {
                    'interface': 'pro-int',
                },
                'pro-rel2': {
                    'interface': 'pro-int',
                },
            },
            'requires': {
                'req-rel': {
                    'interface': 'req-int',
                },
            },
            'peers': {
                'pee-rel': {
                    'interface': 'pee-int',
                },
            },
        }
        rtri = hookenv.relation_to_role_and_interface
        self.assertEqual(rtri('pro-rel'), ('provides', 'pro-int'))
        self.assertEqual(rtri('req-rel'), ('requires', 'req-int'))
        self.assertEqual(rtri('pee-rel'), ('peers', 'pee-int'))

    @patch.object(hookenv, 'metadata')
    def test_role_and_interface_to_relations(self, metadata):
        metadata.return_value = {
            'provides': {
                'pro-rel': {
                    'interface': 'pro-int',
                },
                'pro-rel2': {
                    'interface': 'pro-int',
                },
            },
            'requires': {
                'req-rel': {
                    'interface': 'int',
                },
            },
            'peers': {
                'pee-rel': {
                    'interface': 'int',
                },
            },
        }
        ritr = hookenv.role_and_interface_to_relations
        assertItemsEqual = getattr(self, 'assertItemsEqual', getattr(self, 'assertCountEqual', None))
        assertItemsEqual(ritr('provides', 'pro-int'), ['pro-rel', 'pro-rel2'])
        assertItemsEqual(ritr('requires', 'int'), ['req-rel'])
        assertItemsEqual(ritr('peers', 'int'), ['pee-rel'])

    @patch.object(hookenv, 'metadata')
    def test_interface_to_relations(self, metadata):
        metadata.return_value = {
            'provides': {
                'pro-rel': {
                    'interface': 'pro-int',
                },
                'pro-rel2': {
                    'interface': 'pro-int',
                },
            },
            'requires': {
                'req-rel': {
                    'interface': 'req-int',
                },
            },
            'peers': {
                'pee-rel': {
                    'interface': 'pee-int',
                },
            },
        }
        itr = hookenv.interface_to_relations
        assertItemsEqual = getattr(self, 'assertItemsEqual', getattr(self, 'assertCountEqual', None))
        assertItemsEqual(itr('pro-int'), ['pro-rel', 'pro-rel2'])
        assertItemsEqual(itr('req-int'), ['req-rel'])
        assertItemsEqual(itr('pee-int'), ['pee-rel'])

    def test_action_name(self):
        with patch.dict('os.environ', JUJU_ACTION_NAME='action-jack'):
            self.assertEqual(hookenv.action_name(), 'action-jack')

    def test_action_uuid(self):
        with patch.dict('os.environ', JUJU_ACTION_UUID='action-jack'):
            self.assertEqual(hookenv.action_uuid(), 'action-jack')

    def test_action_tag(self):
        with patch.dict('os.environ', JUJU_ACTION_TAG='action-jack'):
            self.assertEqual(hookenv.action_tag(), 'action-jack')

    @patch('subprocess.check_output')
    def test_storage_list(self, check_output):
        ids = ['data/0', 'data/1', 'data/2']
        check_output.return_value = json.dumps(ids).encode('UTF-8')

        storage_name = 'arbitrary'
        result = hookenv.storage_list(storage_name)

        self.assertEqual(result, ids)
        check_output.assert_called_with(['storage-list', '--format=json',
                                         storage_name])

    @patch('subprocess.check_output')
    def test_storage_list_notexist(self, check_output):
        import errno
        e = OSError()
        e.errno = errno.ENOENT
        check_output.side_effect = e

        result = hookenv.storage_list()

        self.assertEqual(result, [])
        check_output.assert_called_with(['storage-list', '--format=json'])

    @patch('subprocess.check_output')
    def test_storage_get_notexist(self, check_output):
        # storage_get does not catch ENOENT, because there's no reason why you
        # should be calling storage_get except from a storage hook, or with
        # the result of storage_list (which will return [] as shown above).

        import errno
        e = OSError()
        e.errno = errno.ENOENT
        check_output.side_effect = e
        self.assertRaises(OSError, hookenv.storage_get)

    @patch('subprocess.check_output')
    def test_storage_get(self, check_output):
        expect = {
            'location': '/dev/sda',
            'kind': 'block',
        }
        check_output.return_value = json.dumps(expect).encode('UTF-8')

        result = hookenv.storage_get()

        self.assertEqual(result, expect)
        check_output.assert_called_with(['storage-get', '--format=json'])

    @patch('subprocess.check_output')
    def test_storage_get_attr(self, check_output):
        expect = '/dev/sda'
        check_output.return_value = json.dumps(expect).encode('UTF-8')

        attribute = 'location'
        result = hookenv.storage_get(attribute)

        self.assertEqual(result, expect)
        check_output.assert_called_with(['storage-get', '--format=json',
                                         attribute])

    @patch('subprocess.check_output')
    def test_storage_get_with_id(self, check_output):
        expect = {
            'location': '/dev/sda',
            'kind': 'block',
        }
        check_output.return_value = json.dumps(expect).encode('UTF-8')

        storage_id = 'data/0'
        result = hookenv.storage_get(storage_id=storage_id)

        self.assertEqual(result, expect)
        check_output.assert_called_with(['storage-get', '--format=json',
                                         '-s', storage_id])

    @patch('subprocess.check_output')
    def test_network_get_primary(self, check_output):
        """Ensure that network-get is called correctly and output is returned"""
        check_output.return_value = b'192.168.22.1'
        ip = hookenv.network_get_primary_address('mybinding')
        check_output.assert_called_with(
            ['network-get', '--primary-address', 'mybinding'], stderr=-2)
        self.assertEqual(ip, '192.168.22.1')

    @patch('subprocess.check_output')
    def test_network_get_primary_unsupported(self, check_output):
        """Ensure that NotImplementedError is thrown when run on Juju < 2.0"""
        check_output.side_effect = OSError(2, 'network-get')
        self.assertRaises(NotImplementedError, hookenv.network_get_primary_address,
                          'mybinding')

    @patch('subprocess.check_output')
    def test_network_get_primary_no_binding_found(self, check_output):
        """Ensure that NotImplementedError when no binding is found"""
        check_output.side_effect = CalledProcessError(
            1, 'network-get',
            output='no network config found for binding'.encode('UTF-8'))
        self.assertRaises(hookenv.NoNetworkBinding,
                          hookenv.network_get_primary_address,
                          'doesnotexist')
        check_output.assert_called_with(
            ['network-get', '--primary-address', 'doesnotexist'], stderr=-2)

    @patch('subprocess.check_output')
    def test_network_get_primary_other_exception(self, check_output):
        """Ensure that CalledProcessError still thrown when not
        a missing binding"""
        check_output.side_effect = CalledProcessError(
            1, 'network-get',
            output='any other message'.encode('UTF-8'))
        self.assertRaises(CalledProcessError,
                          hookenv.network_get_primary_address,
                          'mybinding')

    @patch('charmhelpers.core.hookenv.juju_version')
    @patch('subprocess.check_output')
    def test_network_get(self, check_output, juju_version):
        """Ensure that network-get is called correctly"""
        juju_version.return_value = '2.2.0'
        check_output.return_value = b'result'
        hookenv.network_get('endpoint')
        check_output.assert_called_with(
            ['network-get', 'endpoint', '--format', 'yaml'], stderr=-2)

    @patch('charmhelpers.core.hookenv.juju_version')
    @patch('subprocess.check_output')
    def test_network_get_primary_required(self, check_output, juju_version):
        """Ensure that NotImplementedError is thrown with Juju < 2.2.0"""
        check_output.return_value = b'result'

        juju_version.return_value = '2.1.4'
        self.assertRaises(NotImplementedError, hookenv.network_get, 'binding')
        juju_version.return_value = '2.2.0'
        self.assertEquals(hookenv.network_get('endpoint'), 'result')

    @patch('charmhelpers.core.hookenv.juju_version')
    @patch('subprocess.check_output')
    def test_network_get_relation_bound(self, check_output, juju_version):
        """Ensure that network-get supports relation context, requires Juju 2.3"""
        juju_version.return_value = '2.3.0'
        check_output.return_value = b'result'
        hookenv.network_get('endpoint', 'db')
        check_output.assert_called_with(
            ['network-get', 'endpoint', '--format', 'yaml', '-r', 'db'],
            stderr=-2)
        juju_version.return_value = '2.2.8'
        self.assertRaises(NotImplementedError, hookenv.network_get, 'endpoint', 'db')

    @patch('charmhelpers.core.hookenv.juju_version')
    @patch('subprocess.check_output')
    def test_network_get_parses_yaml(self, check_output, juju_version):
        """network-get returns loaded YAML output."""
        juju_version.return_value = '2.3.0'
        check_output.return_value = b"""
bind-addresses:
- macaddress: ""
  interfacename: ""
  addresses:
    - address: 10.136.107.33
      cidr: ""
ingress-addresses:
- 10.136.107.33
        """
        ip = hookenv.network_get('mybinding')
        self.assertEqual(len(ip['bind-addresses']), 1)
        self.assertEqual(ip['ingress-addresses'], ['10.136.107.33'])

    @patch('subprocess.check_call')
    def test_add_metric(self, check_call_):
        hookenv.add_metric(flips='1.5', flops='2.1')
        hookenv.add_metric('juju-units=6')
        hookenv.add_metric('foo-bar=3.333', 'baz-quux=8', users='2')
        calls = [
            call(['add-metric', 'flips=1.5', 'flops=2.1']),
            call(['add-metric', 'juju-units=6']),
            call(['add-metric', 'baz-quux=8', 'foo-bar=3.333', 'users=2']),
        ]
        check_call_.assert_has_calls(calls)

    @patch('subprocess.check_call')
    @patch.object(hookenv, 'log')
    def test_add_metric_enoent(self, log, _check_call):
        _check_call.side_effect = OSError(2, 'fail')
        hookenv.add_metric(flips='1')
        log.assert_called_with('add-metric failed: flips=1', level='INFO')

    @patch('charmhelpers.core.hookenv.os')
    def test_meter_status(self, os_):
        os_.environ = {
            'JUJU_METER_STATUS': 'GREEN',
            'JUJU_METER_INFO': 'all good',
        }
        self.assertEqual(hookenv.meter_status(), 'GREEN')
        self.assertEqual(hookenv.meter_info(), 'all good')

    @patch.object(hookenv, 'related_units')
    @patch.object(hookenv, 'relation_ids')
    def test_iter_units_for_relation_name(self, relation_ids, related_units):
        relation_ids.return_value = ['rel:1']
        related_units.return_value = ['unit/0', 'unit/1', 'unit/2']
        expected = [('rel:1', 'unit/0'),
                    ('rel:1', 'unit/1'),
                    ('rel:1', 'unit/2')]
        related_units_data = [
            (u.rid, u.unit)
            for u in hookenv.iter_units_for_relation_name('rel')]
        self.assertEqual(expected, related_units_data)

    @patch.object(hookenv, 'relation_get')
    def test_ingress_address(self, relation_get):
        """Ensure ingress_address returns the ingress-address when available
        and returns the private-address when not.
        """
        _with_ingress = {'egress-subnets': '10.5.0.23/32',
                         'ingress-address': '10.5.0.23',
                         'private-address': '172.16.5.10'}

        _without_ingress = {'private-address': '172.16.5.10'}

        # Return the ingress-address
        relation_get.return_value = _with_ingress
        self.assertEqual(hookenv.ingress_address(rid='test:1', unit='unit/1'),
                         '10.5.0.23')
        relation_get.assert_called_with(rid='test:1', unit='unit/1')
        # Return the private-address
        relation_get.return_value = _without_ingress
        self.assertEqual(hookenv.ingress_address(rid='test:1'),
                         '172.16.5.10')

    @patch.object(hookenv, 'relation_get')
    def test_egress_subnets(self, relation_get):
        """Ensure egress_subnets returns the decoded egress-subnets when available
        and falls back correctly when not.
        """
        d = {'egress-subnets': '10.5.0.23/32,2001::F00F/64',
             'ingress-address': '10.5.0.23',
             'private-address': '2001::D0:F00D'}

        # Return the egress-subnets
        relation_get.return_value = d
        self.assertEqual(hookenv.egress_subnets(rid='test:1', unit='unit/1'),
                         ['10.5.0.23/32', '2001::F00F/64'])
        relation_get.assert_called_with(rid='test:1', unit='unit/1')

        # Return the ingress-address
        del d['egress-subnets']
        self.assertEqual(hookenv.egress_subnets(), ['10.5.0.23/32'])

        # Return the private-address
        del d['ingress-address']
        self.assertEqual(hookenv.egress_subnets(), ['2001::D0:F00D/128'])

    @patch('charmhelpers.core.hookenv.local_unit')
    @patch('charmhelpers.core.hookenv.goal_state')
    @patch('charmhelpers.core.hookenv.has_juju_version')
    def test_unit_doomed(self, has_juju_version, goal_state, local_unit):
        # We need to test for a minimum patch level, or we risk
        # data loss by returning bogus results with Juju 2.4.0
        has_juju_version.return_value = False
        self.assertRaises(NotImplementedError, hookenv.unit_doomed)
        has_juju_version.assertCalledOnceWith("2.4.1")
        has_juju_version.return_value = True

        goal_state.return_value = json.loads('''
                                             {
                                                "units": {
                                                    "postgresql/0": {
                                                        "status": "dying",
                                                        "since": "2018-07-30 10:01:06Z"
                                                    },
                                                    "postgresql/1": {
                                                        "status": "active",
                                                        "since": "2018-07-30 10:22:39Z"
                                                    }
                                                },
                                                "relations": {}
                                             }
                                             ''')
        self.assertTrue(hookenv.unit_doomed('postgresql/0'))   # unit removed, status "dying"
        self.assertFalse(hookenv.unit_doomed('postgresql/1'))  # unit exists, status "active", maybe other states
        self.assertTrue(hookenv.unit_doomed('postgresql/2'))   # unit does not exist

        local_unit.return_value = 'postgresql/0'
        self.assertTrue(hookenv.unit_doomed())

    def test_contains_addr_range(self):
        # Contains cidr
        self.assertTrue(hookenv._contains_range("192.168.1/20"))
        self.assertTrue(hookenv._contains_range("192.168.0/24"))
        self.assertTrue(
            hookenv._contains_range("10.40.50.1,192.168.1/20,10.56.78.9"))
        self.assertTrue(hookenv._contains_range("192.168.22/24"))
        self.assertTrue(hookenv._contains_range("2001:db8::/32"))
        self.assertTrue(hookenv._contains_range("*.foo.com"))
        self.assertTrue(hookenv._contains_range(".foo.com"))
        self.assertTrue(
            hookenv._contains_range("192.168.1.20,.foo.com"))
        self.assertTrue(
            hookenv._contains_range("192.168.1.20,  .foo.com"))
        self.assertTrue(
            hookenv._contains_range("192.168.1.20,*.foo.com"))

        # Doesn't contain cidr
        self.assertFalse(hookenv._contains_range("192.168.1"))
        self.assertFalse(hookenv._contains_range("192.168.145"))
        self.assertFalse(hookenv._contains_range("192.16.14"))



#!/usr/bin/env python

import unittest

from mock import patch, PropertyMock

import charmhelpers.contrib.openstack.keystone as keystone

TO_PATCH = [
    'apt_install',
    "log",
    "ERROR",
    "IdentityServiceContext",
]


class KeystoneTests(unittest.TestCase):
    def setUp(self):
        for m in TO_PATCH:
            setattr(self, m, self._patch(m))

    def _patch(self, method):
        _m = patch('charmhelpers.contrib.openstack.keystone.' + method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        return mock

    def test_get_keystone_manager(self):
        manager = keystone.get_keystone_manager(
            'test-endpoint', 2, token="12345"
        )
        self.assertTrue(isinstance(manager, keystone.KeystoneManager2))

        manager = keystone.get_keystone_manager(
            'test-endpoint', 3, token="12345")

        self.assertTrue(isinstance(manager, keystone.KeystoneManager3))
        self.assertRaises(ValueError, keystone.get_keystone_manager,
                          'test-endpoint', 4, token="12345")

    def test_resolve_sevice_id_v2(self):
        class ServiceList(list):
            def __iter__(self):
                class Service(object):
                    _info = {
                        'type': 'metering',
                        'name': "ceilometer",
                        'id': "uuid-uuid",
                    }
                yield Service()

        manager = keystone.get_keystone_manager('test-endpoint', 2,
                                                token="1234")
        manager.api.services.list = PropertyMock(return_value=ServiceList())
        self.assertTrue(manager.service_exists(service_name="ceilometer",
                                               service_type="metering"))
        self.assertFalse(manager.service_exists(service_name="barbican"))
        self.assertFalse(manager.service_exists(service_name="barbican",
                                                service_type="openstack"))

    def test_resolve_sevice_id_v3(self):
        class ServiceList(list):
            def __iter__(self):
                class Service(object):
                    _info = {
                        'type': 'metering',
                        'name': "ceilometer",
                        'id': "uuid-uuid",
                    }
                yield Service()

        manager = keystone.get_keystone_manager('test-endpoint', 3,
                                                token="12345")
        manager.api.services.list = PropertyMock(return_value=ServiceList())
        self.assertTrue(manager.service_exists(service_name="ceilometer",
                                               service_type="metering"))
        self.assertFalse(manager.service_exists(service_name="barbican"))
        self.assertFalse(manager.service_exists(service_name="barbican",
                                                service_type="openstack"))

    def test_get_api_suffix(self):
        self.assertEquals(keystone.get_api_suffix(2), "v2.0")
        self.assertEquals(keystone.get_api_suffix(3), "v3")

    def test_format_endpoint(self):
        self.assertEquals(keystone.format_endpoint(
            "http", "10.0.0.5", "5000", 2), "http://10.0.0.5:5000/v2.0/")

    def test_get_keystone_manager_from_identity_service_context(self):
        class FakeIdentityServiceV2(object):
            def __call__(self, *args, **kwargs):
                return {
                    "service_protocol": "https",
                    "service_host": "10.5.0.5",
                    "service_port": "5000",
                    "api_version": "2.0",
                    "admin_user": "amdin",
                    "admin_password": "admin",
                    "admin_tenant_name": "admin_tenant"
                }

        self.IdentityServiceContext.return_value = FakeIdentityServiceV2()

        manager = keystone.get_keystone_manager_from_identity_service_context()
        self.assertIsInstance(manager, keystone.KeystoneManager2)

        class FakeIdentityServiceV3(object):
            def __call__(self, *args, **kwargs):
                return {
                    "service_protocol": "https",
                    "service_host": "10.5.0.5",
                    "service_port": "5000",
                    "api_version": "3",
                    "admin_user": "amdin",
                    "admin_password": "admin",
                    "admin_tenant_name": "admin_tenant"
                }

        self.IdentityServiceContext.return_value = FakeIdentityServiceV3()

        manager = keystone.get_keystone_manager_from_identity_service_context()
        self.assertIsInstance(manager, keystone.KeystoneManager3)



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
import os
import six

from mock import call, patch
from unittest import TestCase

from charmhelpers.contrib.hardening import templating
from charmhelpers.contrib.hardening import utils
from charmhelpers.contrib.hardening.audits.file import (
    TemplatedFile,
    FileContentAudit,
)
from charmhelpers.contrib.hardening.ssh.checks import (
    config as ssh_config_check
)
from charmhelpers.contrib.hardening.host.checks import (
    sysctl,
    securetty,
)
from charmhelpers.contrib.hardening.apache.checks import (
    config as apache_config_check
)
from charmhelpers.contrib.hardening.mysql.checks import (
    config as mysql_config_check
)


class TemplatingTestCase(TestCase):

    def setUp(self):
        super(TemplatingTestCase, self).setUp()

        os.environ['JUJU_CHARM_DIR'] = '/tmp'
        self.pathindex = {}
        self.addCleanup(lambda: os.environ.pop('JUJU_CHARM_DIR'))

    def get_renderers(self, audits):
        renderers = []
        for a in audits:
            if issubclass(a.__class__, TemplatedFile):
                renderers.append(a)

        return renderers

    def get_contentcheckers(self, audits):
        contentcheckers = []
        for a in audits:
            if issubclass(a.__class__, FileContentAudit):
                contentcheckers.append(a)

        return contentcheckers

    def render(self, renderers):
        for template in renderers:
            with patch.object(template, 'pre_write', lambda: None):
                with patch.object(template, 'post_write', lambda: None):
                    with patch.object(template, 'run_service_actions'):
                        with patch.object(template, 'save_checksum'):
                            for p in template.paths:
                                template.comply(p)

    def checkcontents(self, contentcheckers):
        for check in contentcheckers:
            if check.path not in self.pathindex:
                continue

            self.assertTrue(check.is_compliant(self.pathindex[check.path]))

    @patch.object(ssh_config_check, 'lsb_release',
                  lambda: {'DISTRIB_CODENAME': 'precise'})
    @patch.object(utils, 'ensure_permissions')
    @patch.object(templating, 'write')
    @patch('charmhelpers.contrib.hardening.audits.file.log')
    @patch.object(templating, 'log', lambda *args, **kwargs: None)
    @patch.object(utils, 'log', lambda *args, **kwargs: None)
    @patch.object(ssh_config_check, 'log', lambda *args, **kwargs: None)
    def test_ssh_config_render_and_check_lt_trusty(self, mock_log, mock_write,
                                                   mock_ensure_permissions):
        audits = ssh_config_check.get_audits()
        contentcheckers = self.get_contentcheckers(audits)
        renderers = self.get_renderers(audits)
        configs = {}

        def write(path, data):
            with tempfile.NamedTemporaryFile(delete=False) as ftmp:
                if os.path.basename(path) == "ssh_config":
                    configs['ssh'] = ftmp.name
                elif os.path.basename(path) == "sshd_config":
                    configs['sshd'] = ftmp.name

                if path in self.pathindex:
                    raise Exception("File already rendered '%s'" % path)

                self.pathindex[path] = ftmp.name
                with open(ftmp.name, 'wb') as fd:
                    fd.write(data)

        mock_write.side_effect = write
        self.render(renderers)
        self.checkcontents(contentcheckers)
        self.assertTrue(mock_write.called)
        args_list = mock_write.call_args_list
        self.assertEqual('/etc/ssh/ssh_config', args_list[0][0][0])
        self.assertEqual('/etc/ssh/sshd_config', args_list[1][0][0])
        self.assertEqual(mock_write.call_count, 2)

        calls = [call("Auditing contents of file '%s'" % configs['ssh'],
                      level='DEBUG'),
                 call('Checked 10 cases and 10 passed', level='DEBUG'),
                 call("Auditing contents of file '%s'" % configs['sshd'],
                      level='DEBUG'),
                 call('Checked 10 cases and 10 passed', level='DEBUG')]
        mock_log.assert_has_calls(calls)

    @patch.object(ssh_config_check, 'lsb_release',
                  lambda: {'DISTRIB_CODENAME': 'trusty'})
    @patch.object(utils, 'ensure_permissions')
    @patch.object(templating, 'write')
    @patch('charmhelpers.contrib.hardening.audits.file.log')
    @patch.object(templating, 'log', lambda *args, **kwargs: None)
    @patch.object(utils, 'log', lambda *args, **kwargs: None)
    @patch.object(ssh_config_check, 'log', lambda *args, **kwargs: None)
    def test_ssh_config_render_and_check_gte_trusty(self, mock_log, mock_write,
                                                    mock_ensure_permissions):
        audits = ssh_config_check.get_audits()
        contentcheckers = self.get_contentcheckers(audits)
        renderers = self.get_renderers(audits)

        def write(path, data):
            with tempfile.NamedTemporaryFile(delete=False) as ftmp:
                if path in self.pathindex:
                    raise Exception("File already rendered '%s'" % path)

                self.pathindex[path] = ftmp.name
                with open(ftmp.name, 'wb') as fd:
                    fd.write(data)

        mock_write.side_effect = write
        self.render(renderers)
        self.checkcontents(contentcheckers)
        self.assertTrue(mock_write.called)
        args_list = mock_write.call_args_list
        self.assertEqual('/etc/ssh/ssh_config', args_list[0][0][0])
        self.assertEqual('/etc/ssh/sshd_config', args_list[1][0][0])
        self.assertEqual(mock_write.call_count, 2)

        mock_log.assert_has_calls([call('Checked 9 cases and 9 passed',
                                        level='DEBUG')])

    @patch.object(utils, 'ensure_permissions')
    @patch.object(templating, 'write')
    @patch.object(sysctl, 'log', lambda *args, **kwargs: None)
    @patch.object(templating, 'log', lambda *args, **kwargs: None)
    @patch.object(utils, 'log', lambda *args, **kwargs: None)
    def test_os_sysctl_and_check(self, mock_write, mock_ensure_permissions):
        audits = sysctl.get_audits()
        contentcheckers = self.get_contentcheckers(audits)
        renderers = self.get_renderers(audits)

        def write(path, data):
            if path in self.pathindex:
                raise Exception("File already rendered '%s'" % path)

            with tempfile.NamedTemporaryFile(delete=False) as ftmp:
                self.pathindex[path] = ftmp.name
                with open(ftmp.name, 'wb') as fd:
                    fd.write(data)

        mock_write.side_effect = write
        self.render(renderers)
        self.checkcontents(contentcheckers)
        self.assertTrue(mock_write.called)
        args_list = mock_write.call_args_list
        self.assertEqual('/etc/sysctl.d/99-juju-hardening.conf',
                         args_list[0][0][0])
        self.assertEqual(mock_write.call_count, 1)

    @patch.object(utils, 'ensure_permissions')
    @patch.object(templating, 'write')
    @patch.object(sysctl, 'log', lambda *args, **kwargs: None)
    @patch.object(templating, 'log', lambda *args, **kwargs: None)
    @patch.object(utils, 'log', lambda *args, **kwargs: None)
    def test_os_securetty_and_check(self, mock_write, mock_ensure_permissions):
        audits = securetty.get_audits()
        contentcheckers = self.get_contentcheckers(audits)
        renderers = self.get_renderers(audits)

        def write(path, data):
            if path in self.pathindex:
                raise Exception("File already rendered '%s'" % path)

            with tempfile.NamedTemporaryFile(delete=False) as ftmp:
                self.pathindex[path] = ftmp.name
                with open(ftmp.name, 'wb') as fd:
                    fd.write(data)

        mock_write.side_effect = write
        self.render(renderers)
        self.checkcontents(contentcheckers)
        self.assertTrue(mock_write.called)
        args_list = mock_write.call_args_list
        self.assertEqual('/etc/securetty', args_list[0][0][0])
        self.assertEqual(mock_write.call_count, 1)

    @patch.object(apache_config_check.utils, 'get_settings', lambda x: {
        'common': {'apache_dir': '/tmp/foo'},
        'hardening': {
            'allowed_http_methods': {'GOGETEM'},
            'modules_to_disable': {'modfoo'},
            'traceenable': 'off',
            'servertokens': 'Prod',
            'honor_cipher_order': 'on',
            'cipher_suite': 'ALL:+MEDIUM:+HIGH:!LOW:!MD5:!RC4:!eNULL:!aNULL:!3DES'
        }
    })
    @patch('charmhelpers.contrib.hardening.audits.file.os.path.exists',
           lambda *a, **kwa: True)
    @patch.object(apache_config_check, 'subprocess')
    @patch.object(utils, 'ensure_permissions')
    @patch.object(templating, 'write')
    @patch.object(templating, 'log', lambda *args, **kwargs: None)
    @patch.object(utils, 'log', lambda *args, **kwargs: None)
    def test_apache_conf_and_check(self, mock_write, mock_ensure_permissions,
                                   mock_subprocess):
        mock_subprocess.call.return_value = 0
        apache_version = b"""Server version: Apache/2.4.7 (Ubuntu)
        Server built:   Jan 14 2016 17:45:23
        """
        mock_subprocess.check_output.return_value = apache_version
        audits = apache_config_check.get_audits()
        contentcheckers = self.get_contentcheckers(audits)
        renderers = self.get_renderers(audits)

        def write(path, data):
            if path in self.pathindex:
                raise Exception("File already rendered '%s'" % path)

            with tempfile.NamedTemporaryFile(delete=False) as ftmp:
                self.pathindex[path] = ftmp.name
                with open(ftmp.name, 'wb') as fd:
                    fd.write(data)

        mock_write.side_effect = write
        self.render(renderers)
        self.checkcontents(contentcheckers)
        self.assertTrue(mock_write.called)
        args_list = mock_write.call_args_list
        self.assertEqual('/tmp/foo/mods-available/alias.conf',
                         args_list[0][0][0])
        self.assertEqual(mock_write.call_count, 2)

    @patch.object(apache_config_check.utils, 'get_settings', lambda x: {
        'security': {},
        'hardening': {
            'mysql-conf': '/tmp/foo/mysql.cnf',
            'hardening-conf': '/tmp/foo/conf.d/hardening.cnf'
        }
    })
    @patch('charmhelpers.contrib.hardening.audits.file.os.path.exists',
           lambda *a, **kwa: True)
    @patch.object(utils, 'ensure_permissions')
    @patch.object(templating, 'write')
    @patch.object(mysql_config_check.subprocess, 'call',
                  lambda *args, **kwargs: 0)
    @patch.object(templating, 'log', lambda *args, **kwargs: None)
    @patch.object(utils, 'log', lambda *args, **kwargs: None)
    def test_mysql_conf_and_check(self, mock_write, mock_ensure_permissions):
        audits = mysql_config_check.get_audits()
        contentcheckers = self.get_contentcheckers(audits)
        renderers = self.get_renderers(audits)

        def write(path, data):
            if path in self.pathindex:
                raise Exception("File already rendered '%s'" % path)

            with tempfile.NamedTemporaryFile(delete=False) as ftmp:
                self.pathindex[path] = ftmp.name
                with open(ftmp.name, 'wb') as fd:
                    fd.write(data)

        mock_write.side_effect = write
        self.render(renderers)
        self.checkcontents(contentcheckers)
        self.assertTrue(mock_write.called)
        args_list = mock_write.call_args_list
        self.assertEqual('/tmp/foo/conf.d/hardening.cnf',
                         args_list[0][0][0])
        self.assertEqual(mock_write.call_count, 1)

    def tearDown(self):
        # Cleanup
        for path in six.itervalues(self.pathindex):
            os.remove(path)

        super(TemplatingTestCase, self).tearDown()



# Copyright 2019 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import configparser
import glob
import os.path
import subprocess

from charmhelpers.contrib.openstack.audits import (
    audit,
    AuditType,
    # filters
    is_audit_type,
    it_has_config,
)

from charmhelpers.core.hookenv import (
    cached,
)


FILE_ASSERTIONS = {
    'barbican': {
        # From security guide
        '/etc/barbican/barbican.conf': {'group': 'barbican', 'mode': '640'},
        '/etc/barbican/barbican-api-paste.ini':
            {'group': 'barbican', 'mode': '640'},
        '/etc/barbican/policy.json': {'group': 'barbican', 'mode': '640'},
    },
    'ceph-mon': {
        '/var/lib/charm/ceph-mon/ceph.conf':
            {'owner': 'root', 'group': 'root', 'mode': '644'},
        '/etc/ceph/ceph.client.admin.keyring':
            {'owner': 'ceph', 'group': 'ceph'},
        '/etc/ceph/rbdmap': {'mode': '644'},
        '/var/lib/ceph': {'owner': 'ceph', 'group': 'ceph', 'mode': '750'},
        '/var/lib/ceph/bootstrap-*/ceph.keyring':
            {'owner': 'ceph', 'group': 'ceph', 'mode': '600'}
    },
    'ceph-osd': {
        '/var/lib/charm/ceph-osd/ceph.conf':
            {'owner': 'ceph', 'group': 'ceph', 'mode': '644'},
        '/var/lib/ceph': {'owner': 'ceph', 'group': 'ceph', 'mode': '750'},
        '/var/lib/ceph/*': {'owner': 'ceph', 'group': 'ceph', 'mode': '755'},
        '/var/lib/ceph/bootstrap-*/ceph.keyring':
            {'owner': 'ceph', 'group': 'ceph', 'mode': '600'},
        '/var/lib/ceph/radosgw':
            {'owner': 'ceph', 'group': 'ceph', 'mode': '755'},
    },
    'cinder': {
        # From security guide
        '/etc/cinder/cinder.conf': {'group': 'cinder', 'mode': '640'},
        '/etc/cinder/api-paste.conf': {'group': 'cinder', 'mode': '640'},
        '/etc/cinder/rootwrap.conf': {'group': 'cinder', 'mode': '640'},
    },
    'glance': {
        # From security guide
        '/etc/glance/glance-api-paste.ini': {'group': 'glance', 'mode': '640'},
        '/etc/glance/glance-api.conf': {'group': 'glance', 'mode': '640'},
        '/etc/glance/glance-cache.conf': {'group': 'glance', 'mode': '640'},
        '/etc/glance/glance-manage.conf': {'group': 'glance', 'mode': '640'},
        '/etc/glance/glance-registry-paste.ini':
            {'group': 'glance', 'mode': '640'},
        '/etc/glance/glance-registry.conf': {'group': 'glance', 'mode': '640'},
        '/etc/glance/glance-scrubber.conf': {'group': 'glance', 'mode': '640'},
        '/etc/glance/glance-swift-store.conf':
            {'group': 'glance', 'mode': '640'},
        '/etc/glance/policy.json': {'group': 'glance', 'mode': '640'},
        '/etc/glance/schema-image.json': {'group': 'glance', 'mode': '640'},
        '/etc/glance/schema.json': {'group': 'glance', 'mode': '640'},
    },
    'keystone': {
        # From security guide
        '/etc/keystone/keystone.conf': {'group': 'keystone', 'mode': '640'},
        '/etc/keystone/keystone-paste.ini':
            {'group': 'keystone', 'mode': '640'},
        '/etc/keystone/policy.json': {'group': 'keystone', 'mode': '640'},
        '/etc/keystone/logging.conf': {'group': 'keystone', 'mode': '640'},
        '/etc/keystone/ssl/certs/signing_cert.pem':
            {'group': 'keystone', 'mode': '640'},
        '/etc/keystone/ssl/private/signing_key.pem':
            {'group': 'keystone', 'mode': '640'},
        '/etc/keystone/ssl/certs/ca.pem': {'group': 'keystone', 'mode': '640'},
    },
    'manilla': {
        # From security guide
        '/etc/manila/manila.conf': {'group': 'manilla', 'mode': '640'},
        '/etc/manila/api-paste.ini': {'group': 'manilla', 'mode': '640'},
        '/etc/manila/policy.json': {'group': 'manilla', 'mode': '640'},
        '/etc/manila/rootwrap.conf': {'group': 'manilla', 'mode': '640'},
    },
    'neutron-gateway': {
        '/etc/neutron/neutron.conf': {'group': 'neutron', 'mode': '640'},
        '/etc/neutron/rootwrap.conf': {'mode': '640'},
        '/etc/neutron/rootwrap.d': {'mode': '755'},
        '/etc/neutron/*': {'group': 'neutron', 'mode': '644'},
    },
    'neutron-api': {
        # From security guide
        '/etc/neutron/neutron.conf': {'group': 'neutron', 'mode': '640'},
        '/etc/nova/api-paste.ini': {'group': 'neutron', 'mode': '640'},
        '/etc/neutron/rootwrap.conf': {'group': 'neutron', 'mode': '640'},
        # Additional validations
        '/etc/neutron/rootwrap.d': {'mode': '755'},
        '/etc/neutron/neutron_lbaas.conf': {'mode': '644'},
        '/etc/neutron/neutron_vpnaas.conf': {'mode': '644'},
        '/etc/neutron/*': {'group': 'neutron', 'mode': '644'},
    },
    'nova-cloud-controller': {
        # From security guide
        '/etc/nova/api-paste.ini': {'group': 'nova', 'mode': '640'},
        '/etc/nova/nova.conf': {'group': 'nova', 'mode': '750'},
        '/etc/nova/*': {'group': 'nova', 'mode': '640'},
        # Additional validations
        '/etc/nova/logging.conf': {'group': 'nova', 'mode': '640'},
    },
    'nova-compute': {
        # From security guide
        '/etc/nova/nova.conf': {'group': 'nova', 'mode': '640'},
        '/etc/nova/api-paste.ini': {'group': 'nova', 'mode': '640'},
        '/etc/nova/rootwrap.conf': {'group': 'nova', 'mode': '640'},
        # Additional Validations
        '/etc/nova/nova-compute.conf': {'group': 'nova', 'mode': '640'},
        '/etc/nova/logging.conf': {'group': 'nova', 'mode': '640'},
        '/etc/nova/nm.conf': {'mode': '644'},
        '/etc/nova/*': {'group': 'nova', 'mode': '640'},
    },
    'openstack-dashboard': {
        # From security guide
        '/etc/openstack-dashboard/local_settings.py':
            {'group': 'horizon', 'mode': '640'},
    },
}

Ownership = collections.namedtuple('Ownership', 'owner group mode')


@cached
def _stat(file):
    """
    Get the Ownership information from a file.

    :param file: The path to a file to stat
    :type file: str
    :returns: owner, group, and mode of the specified file
    :rtype: Ownership
    :raises subprocess.CalledProcessError: If the underlying stat fails
    """
    out = subprocess.check_output(
        ['stat', '-c', '%U %G %a', file]).decode('utf-8')
    return Ownership(*out.strip().split(' '))


@cached
def _config_ini(path):
    """
    Parse an ini file

    :param path: The path to a file to parse
    :type file: str
    :returns: Configuration contained in path
    :rtype: Dict
    """
    conf = configparser.ConfigParser()
    conf.read(path)
    return dict(conf)


def _validate_file_ownership(owner, group, file_name):
    """
    Validate that a specified file is owned by `owner:group`.

    :param owner: Name of the owner
    :type owner: str
    :param group: Name of the group
    :type group: str
    :param file_name: Path to the file to verify
    :type file_name: str
    """
    try:
        ownership = _stat(file_name)
    except subprocess.CalledProcessError as e:
        print("Error reading file: {}".format(e))
        assert False, "Specified file does not exist: {}".format(file_name)
    assert owner == ownership.owner, \
        "{} has an incorrect owner: {} should be {}".format(
            file_name, ownership.owner, owner)
    assert group == ownership.group, \
        "{} has an incorrect group: {} should be {}".format(
            file_name, ownership.group, group)
    print("Validate ownership of {}: PASS".format(file_name))


def _validate_file_mode(mode, file_name):
    """
    Validate that a specified file has the specified permissions.

    :param mode: file mode that is desires
    :type owner: str
    :param file_name: Path to the file to verify
    :type file_name: str
    """
    try:
        ownership = _stat(file_name)
    except subprocess.CalledProcessError as e:
        print("Error reading file: {}".format(e))
        assert False, "Specified file does not exist: {}".format(file_name)
    assert mode == ownership.mode, \
        "{} has an incorrect mode: {} should be {}".format(
            file_name, ownership.mode, mode)
    print("Validate mode of {}: PASS".format(file_name))


@cached
def _config_section(config, section):
    """Read the configuration file and return a section."""
    path = os.path.join(config.get('config_path'), config.get('config_file'))
    conf = _config_ini(path)
    return conf.get(section)


@audit(is_audit_type(AuditType.OpenStackSecurityGuide),
       it_has_config('files'))
def validate_file_ownership(config):
    """Verify that configuration files are owned by the correct user/group."""
    files = config.get('files', {})
    for file_name, options in files.items():
        for key in options.keys():
            if key not in ["owner", "group", "mode"]:
                raise RuntimeError(
                    "Invalid ownership configuration: {}".format(key))
        owner = options.get('owner', config.get('owner', 'root'))
        group = options.get('group', config.get('group', 'root'))
        if '*' in file_name:
            for file in glob.glob(file_name):
                if file not in files.keys():
                    if os.path.isfile(file):
                        _validate_file_ownership(owner, group, file)
        else:
            if os.path.isfile(file_name):
                _validate_file_ownership(owner, group, file_name)


@audit(is_audit_type(AuditType.OpenStackSecurityGuide),
       it_has_config('files'))
def validate_file_permissions(config):
    """Verify that permissions on configuration files are secure enough."""
    files = config.get('files', {})
    for file_name, options in files.items():
        for key in options.keys():
            if key not in ["owner", "group", "mode"]:
                raise RuntimeError(
                    "Invalid ownership configuration: {}".format(key))
        mode = options.get('mode', config.get('permissions', '600'))
        if '*' in file_name:
            for file in glob.glob(file_name):
                if file not in files.keys():
                    if os.path.isfile(file):
                        _validate_file_mode(mode, file)
        else:
            if os.path.isfile(file_name):
                _validate_file_mode(mode, file_name)


@audit(is_audit_type(AuditType.OpenStackSecurityGuide))
def validate_uses_keystone(audit_options):
    """Validate that the service uses Keystone for authentication."""
    section = _config_section(audit_options, 'DEFAULT')
    assert section is not None, "Missing section 'DEFAULT'"
    assert section.get('auth_strategy') == "keystone", \
        "Application is not using Keystone"


@audit(is_audit_type(AuditType.OpenStackSecurityGuide))
def validate_uses_tls_for_keystone(audit_options):
    """Verify that TLS is used to communicate with Keystone."""
    section = _config_section(audit_options, 'keystone_authtoken')
    assert section is not None, "Missing section 'keystone_authtoken'"
    assert not section.get('insecure') and \
        "https://" in section.get("auth_uri"), \
        "TLS is not used for Keystone"


@audit(is_audit_type(AuditType.OpenStackSecurityGuide))
def validate_uses_tls_for_glance(audit_options):
    """Verify that TLS is used to communicate with Glance."""
    section = _config_section(audit_options, 'glance')
    assert section is not None, "Missing section 'glance'"
    assert not section.get('insecure') and \
        "https://" in section.get("api_servers"), \
        "TLS is not used for Glance"



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import logging
import os
import re
import socket
import subprocess
import sys
import time
import uuid

import amulet
import distro_info
import six
from six.moves import configparser
if six.PY3:
    from urllib import parse as urlparse
else:
    import urlparse


class AmuletUtils(object):
    """Amulet utilities.

       This class provides common utility functions that are used by Amulet
       tests.
       """

    def __init__(self, log_level=logging.ERROR):
        self.log = self.get_logger(level=log_level)
        self.ubuntu_releases = self.get_ubuntu_releases()

    def get_logger(self, name="amulet-logger", level=logging.DEBUG):
        """Get a logger object that will log to stdout."""
        log = logging
        logger = log.getLogger(name)
        fmt = log.Formatter("%(asctime)s %(funcName)s "
                            "%(levelname)s: %(message)s")

        handler = log.StreamHandler(stream=sys.stdout)
        handler.setLevel(level)
        handler.setFormatter(fmt)

        logger.addHandler(handler)
        logger.setLevel(level)

        return logger

    def valid_ip(self, ip):
        if re.match(r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$", ip):
            return True
        else:
            return False

    def valid_url(self, url):
        p = re.compile(
            r'^(?:http|ftp)s?://'
            r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # noqa
            r'localhost|'
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'
            r'(?::\d+)?'
            r'(?:/?|[/?]\S+)$',
            re.IGNORECASE)
        if p.match(url):
            return True
        else:
            return False

    def get_ubuntu_release_from_sentry(self, sentry_unit):
        """Get Ubuntu release codename from sentry unit.

        :param sentry_unit: amulet sentry/service unit pointer
        :returns: list of strings - release codename, failure message
        """
        msg = None
        cmd = 'lsb_release -cs'
        release, code = sentry_unit.run(cmd)
        if code == 0:
            self.log.debug('{} lsb_release: {}'.format(
                sentry_unit.info['unit_name'], release))
        else:
            msg = ('{} `{}` returned {} '
                   '{}'.format(sentry_unit.info['unit_name'],
                               cmd, release, code))
        if release not in self.ubuntu_releases:
            msg = ("Release ({}) not found in Ubuntu releases "
                   "({})".format(release, self.ubuntu_releases))
        return release, msg

    def validate_services(self, commands):
        """Validate that lists of commands succeed on service units.  Can be
           used to verify system services are running on the corresponding
           service units.

        :param commands: dict with sentry keys and arbitrary command list vals
        :returns: None if successful, Failure string message otherwise
        """
        self.log.debug('Checking status of system services...')

        # /!\ DEPRECATION WARNING (beisner):
        # New and existing tests should be rewritten to use
        # validate_services_by_name() as it is aware of init systems.
        self.log.warn('DEPRECATION WARNING:  use '
                      'validate_services_by_name instead of validate_services '
                      'due to init system differences.')

        for k, v in six.iteritems(commands):
            for cmd in v:
                output, code = k.run(cmd)
                self.log.debug('{} `{}` returned '
                               '{}'.format(k.info['unit_name'],
                                           cmd, code))
                if code != 0:
                    return "command `{}` returned {}".format(cmd, str(code))
        return None

    def validate_services_by_name(self, sentry_services):
        """Validate system service status by service name, automatically
           detecting init system based on Ubuntu release codename.

        :param sentry_services: dict with sentry keys and svc list values
        :returns: None if successful, Failure string message otherwise
        """
        self.log.debug('Checking status of system services...')

        # Point at which systemd became a thing
        systemd_switch = self.ubuntu_releases.index('vivid')

        for sentry_unit, services_list in six.iteritems(sentry_services):
            # Get lsb_release codename from unit
            release, ret = self.get_ubuntu_release_from_sentry(sentry_unit)
            if ret:
                return ret

            for service_name in services_list:
                if (self.ubuntu_releases.index(release) >= systemd_switch or
                        service_name in ['rabbitmq-server', 'apache2',
                                         'memcached']):
                    # init is systemd (or regular sysv)
                    cmd = 'sudo service {} status'.format(service_name)
                    output, code = sentry_unit.run(cmd)
                    service_running = code == 0
                elif self.ubuntu_releases.index(release) < systemd_switch:
                    # init is upstart
                    cmd = 'sudo status {}'.format(service_name)
                    output, code = sentry_unit.run(cmd)
                    service_running = code == 0 and "start/running" in output

                self.log.debug('{} `{}` returned '
                               '{}'.format(sentry_unit.info['unit_name'],
                                           cmd, code))
                if not service_running:
                    return u"command `{}` returned {} {}".format(
                        cmd, output, str(code))
        return None

    def _get_config(self, unit, filename):
        """Get a ConfigParser object for parsing a unit's config file."""
        file_contents = unit.file_contents(filename)

        # NOTE(beisner):  by default, ConfigParser does not handle options
        # with no value, such as the flags used in the mysql my.cnf file.
        # https://bugs.python.org/issue7005
        config = configparser.ConfigParser(allow_no_value=True)
        config.readfp(io.StringIO(file_contents))
        return config

    def validate_config_data(self, sentry_unit, config_file, section,
                             expected):
        """Validate config file data.

           Verify that the specified section of the config file contains
           the expected option key:value pairs.

           Compare expected dictionary data vs actual dictionary data.
           The values in the 'expected' dictionary can be strings, bools, ints,
           longs, or can be a function that evaluates a variable and returns a
           bool.
           """
        self.log.debug('Validating config file data ({} in {} on {})'
                       '...'.format(section, config_file,
                                    sentry_unit.info['unit_name']))
        config = self._get_config(sentry_unit, config_file)

        if section != 'DEFAULT' and not config.has_section(section):
            return "section [{}] does not exist".format(section)

        for k in expected.keys():
            if not config.has_option(section, k):
                return "section [{}] is missing option {}".format(section, k)

            actual = config.get(section, k)
            v = expected[k]
            if (isinstance(v, six.string_types) or
                    isinstance(v, bool) or
                    isinstance(v, six.integer_types)):
                # handle explicit values
                if actual != v:
                    return "section [{}] {}:{} != expected {}:{}".format(
                           section, k, actual, k, expected[k])
            # handle function pointers, such as not_null or valid_ip
            elif not v(actual):
                return "section [{}] {}:{} != expected {}:{}".format(
                       section, k, actual, k, expected[k])
        return None

    def _validate_dict_data(self, expected, actual):
        """Validate dictionary data.

           Compare expected dictionary data vs actual dictionary data.
           The values in the 'expected' dictionary can be strings, bools, ints,
           longs, or can be a function that evaluates a variable and returns a
           bool.
           """
        self.log.debug('actual: {}'.format(repr(actual)))
        self.log.debug('expected: {}'.format(repr(expected)))

        for k, v in six.iteritems(expected):
            if k in actual:
                if (isinstance(v, six.string_types) or
                        isinstance(v, bool) or
                        isinstance(v, six.integer_types)):
                    # handle explicit values
                    if v != actual[k]:
                        return "{}:{}".format(k, actual[k])
                # handle function pointers, such as not_null or valid_ip
                elif not v(actual[k]):
                    return "{}:{}".format(k, actual[k])
            else:
                return "key '{}' does not exist".format(k)
        return None

    def validate_relation_data(self, sentry_unit, relation, expected):
        """Validate actual relation data based on expected relation data."""
        actual = sentry_unit.relation(relation[0], relation[1])
        return self._validate_dict_data(expected, actual)

    def _validate_list_data(self, expected, actual):
        """Compare expected list vs actual list data."""
        for e in expected:
            if e not in actual:
                return "expected item {} not found in actual list".format(e)
        return None

    def not_null(self, string):
        if string is not None:
            return True
        else:
            return False

    def _get_file_mtime(self, sentry_unit, filename):
        """Get last modification time of file."""
        return sentry_unit.file_stat(filename)['mtime']

    def _get_dir_mtime(self, sentry_unit, directory):
        """Get last modification time of directory."""
        return sentry_unit.directory_stat(directory)['mtime']

    def _get_proc_start_time(self, sentry_unit, service, pgrep_full=None):
        """Get start time of a process based on the last modification time
           of the /proc/pid directory.

        :sentry_unit:  The sentry unit to check for the service on
        :service:  service name to look for in process table
        :pgrep_full:  [Deprecated] Use full command line search mode with pgrep
        :returns:  epoch time of service process start
        :param commands:  list of bash commands
        :param sentry_units:  list of sentry unit pointers
        :returns:  None if successful; Failure message otherwise
        """
        if pgrep_full is not None:
            # /!\ DEPRECATION WARNING (beisner):
            # No longer implemented, as pidof is now used instead of pgrep.
            # https://bugs.launchpad.net/charm-helpers/+bug/1474030
            self.log.warn('DEPRECATION WARNING:  pgrep_full bool is no '
                          'longer implemented re: lp 1474030.')

        pid_list = self.get_process_id_list(sentry_unit, service)
        pid = pid_list[0]
        proc_dir = '/proc/{}'.format(pid)
        self.log.debug('Pid for {} on {}: {}'.format(
            service, sentry_unit.info['unit_name'], pid))

        return self._get_dir_mtime(sentry_unit, proc_dir)

    def service_restarted(self, sentry_unit, service, filename,
                          pgrep_full=None, sleep_time=20):
        """Check if service was restarted.

           Compare a service's start time vs a file's last modification time
           (such as a config file for that service) to determine if the service
           has been restarted.
           """
        # /!\ DEPRECATION WARNING (beisner):
        # This method is prone to races in that no before-time is known.
        # Use validate_service_config_changed instead.

        # NOTE(beisner) pgrep_full is no longer implemented, as pidof is now
        # used instead of pgrep.  pgrep_full is still passed through to ensure
        # deprecation WARNS.  lp1474030
        self.log.warn('DEPRECATION WARNING:  use '
                      'validate_service_config_changed instead of '
                      'service_restarted due to known races.')

        time.sleep(sleep_time)
        if (self._get_proc_start_time(sentry_unit, service, pgrep_full) >=
                self._get_file_mtime(sentry_unit, filename)):
            return True
        else:
            return False

    def service_restarted_since(self, sentry_unit, mtime, service,
                                pgrep_full=None, sleep_time=20,
                                retry_count=30, retry_sleep_time=10):
        """Check if service was been started after a given time.

        Args:
          sentry_unit (sentry): The sentry unit to check for the service on
          mtime (float): The epoch time to check against
          service (string): service name to look for in process table
          pgrep_full: [Deprecated] Use full command line search mode with pgrep
          sleep_time (int): Initial sleep time (s) before looking for file
          retry_sleep_time (int): Time (s) to sleep between retries
          retry_count (int): If file is not found, how many times to retry

        Returns:
          bool: True if service found and its start time it newer than mtime,
                False if service is older than mtime or if service was
                not found.
        """
        # NOTE(beisner) pgrep_full is no longer implemented, as pidof is now
        # used instead of pgrep.  pgrep_full is still passed through to ensure
        # deprecation WARNS.  lp1474030

        unit_name = sentry_unit.info['unit_name']
        self.log.debug('Checking that %s service restarted since %s on '
                       '%s' % (service, mtime, unit_name))
        time.sleep(sleep_time)
        proc_start_time = None
        tries = 0
        while tries <= retry_count and not proc_start_time:
            try:
                proc_start_time = self._get_proc_start_time(sentry_unit,
                                                            service,
                                                            pgrep_full)
                self.log.debug('Attempt {} to get {} proc start time on {} '
                               'OK'.format(tries, service, unit_name))
            except IOError as e:
                # NOTE(beisner) - race avoidance, proc may not exist yet.
                # https://bugs.launchpad.net/charm-helpers/+bug/1474030
                self.log.debug('Attempt {} to get {} proc start time on {} '
                               'failed\n{}'.format(tries, service,
                                                   unit_name, e))
                time.sleep(retry_sleep_time)
                tries += 1

        if not proc_start_time:
            self.log.warn('No proc start time found, assuming service did '
                          'not start')
            return False
        if proc_start_time >= mtime:
            self.log.debug('Proc start time is newer than provided mtime'
                           '(%s >= %s) on %s (OK)' % (proc_start_time,
                                                      mtime, unit_name))
            return True
        else:
            self.log.warn('Proc start time (%s) is older than provided mtime '
                          '(%s) on %s, service did not '
                          'restart' % (proc_start_time, mtime, unit_name))
            return False

    def config_updated_since(self, sentry_unit, filename, mtime,
                             sleep_time=20, retry_count=30,
                             retry_sleep_time=10):
        """Check if file was modified after a given time.

        Args:
          sentry_unit (sentry): The sentry unit to check the file mtime on
          filename (string): The file to check mtime of
          mtime (float): The epoch time to check against
          sleep_time (int): Initial sleep time (s) before looking for file
          retry_sleep_time (int): Time (s) to sleep between retries
          retry_count (int): If file is not found, how many times to retry

        Returns:
          bool: True if file was modified more recently than mtime, False if
                file was modified before mtime, or if file not found.
        """
        unit_name = sentry_unit.info['unit_name']
        self.log.debug('Checking that %s updated since %s on '
                       '%s' % (filename, mtime, unit_name))
        time.sleep(sleep_time)
        file_mtime = None
        tries = 0
        while tries <= retry_count and not file_mtime:
            try:
                file_mtime = self._get_file_mtime(sentry_unit, filename)
                self.log.debug('Attempt {} to get {} file mtime on {} '
                               'OK'.format(tries, filename, unit_name))
            except IOError as e:
                # NOTE(beisner) - race avoidance, file may not exist yet.
                # https://bugs.launchpad.net/charm-helpers/+bug/1474030
                self.log.debug('Attempt {} to get {} file mtime on {} '
                               'failed\n{}'.format(tries, filename,
                                                   unit_name, e))
                time.sleep(retry_sleep_time)
                tries += 1

        if not file_mtime:
            self.log.warn('Could not determine file mtime, assuming '
                          'file does not exist')
            return False

        if file_mtime >= mtime:
            self.log.debug('File mtime is newer than provided mtime '
                           '(%s >= %s) on %s (OK)' % (file_mtime,
                                                      mtime, unit_name))
            return True
        else:
            self.log.warn('File mtime is older than provided mtime'
                          '(%s < on %s) on %s' % (file_mtime,
                                                  mtime, unit_name))
            return False

    def validate_service_config_changed(self, sentry_unit, mtime, service,
                                        filename, pgrep_full=None,
                                        sleep_time=20, retry_count=30,
                                        retry_sleep_time=10):
        """Check service and file were updated after mtime

        Args:
          sentry_unit (sentry): The sentry unit to check for the service on
          mtime (float): The epoch time to check against
          service (string): service name to look for in process table
          filename (string): The file to check mtime of
          pgrep_full: [Deprecated] Use full command line search mode with pgrep
          sleep_time (int): Initial sleep in seconds to pass to test helpers
          retry_count (int): If service is not found, how many times to retry
          retry_sleep_time (int): Time in seconds to wait between retries

        Typical Usage:
            u = OpenStackAmuletUtils(ERROR)
            ...
            mtime = u.get_sentry_time(self.cinder_sentry)
            self.d.configure('cinder', {'verbose': 'True', 'debug': 'True'})
            if not u.validate_service_config_changed(self.cinder_sentry,
                                                     mtime,
                                                     'cinder-api',
                                                     '/etc/cinder/cinder.conf')
                amulet.raise_status(amulet.FAIL, msg='update failed')
        Returns:
          bool: True if both service and file where updated/restarted after
                mtime, False if service is older than mtime or if service was
                not found or if filename was modified before mtime.
        """

        # NOTE(beisner) pgrep_full is no longer implemented, as pidof is now
        # used instead of pgrep.  pgrep_full is still passed through to ensure
        # deprecation WARNS.  lp1474030

        service_restart = self.service_restarted_since(
            sentry_unit, mtime,
            service,
            pgrep_full=pgrep_full,
            sleep_time=sleep_time,
            retry_count=retry_count,
            retry_sleep_time=retry_sleep_time)

        config_update = self.config_updated_since(
            sentry_unit,
            filename,
            mtime,
            sleep_time=sleep_time,
            retry_count=retry_count,
            retry_sleep_time=retry_sleep_time)

        return service_restart and config_update

    def get_sentry_time(self, sentry_unit):
        """Return current epoch time on a sentry"""
        cmd = "date +'%s'"
        return float(sentry_unit.run(cmd)[0])

    def relation_error(self, name, data):
        return 'unexpected relation data in {} - {}'.format(name, data)

    def endpoint_error(self, name, data):
        return 'unexpected endpoint data in {} - {}'.format(name, data)

    def get_ubuntu_releases(self):
        """Return a list of all Ubuntu releases in order of release."""
        _d = distro_info.UbuntuDistroInfo()
        _release_list = _d.all
        return _release_list

    def file_to_url(self, file_rel_path):
        """Convert a relative file path to a file URL."""
        _abs_path = os.path.abspath(file_rel_path)
        return urlparse.urlparse(_abs_path, scheme='file').geturl()

    def check_commands_on_units(self, commands, sentry_units):
        """Check that all commands in a list exit zero on all
        sentry units in a list.

        :param commands:  list of bash commands
        :param sentry_units:  list of sentry unit pointers
        :returns: None if successful; Failure message otherwise
        """
        self.log.debug('Checking exit codes for {} commands on {} '
                       'sentry units...'.format(len(commands),
                                                len(sentry_units)))
        for sentry_unit in sentry_units:
            for cmd in commands:
                output, code = sentry_unit.run(cmd)
                if code == 0:
                    self.log.debug('{} `{}` returned {} '
                                   '(OK)'.format(sentry_unit.info['unit_name'],
                                                 cmd, code))
                else:
                    return ('{} `{}` returned {} '
                            '{}'.format(sentry_unit.info['unit_name'],
                                        cmd, code, output))
        return None

    def get_process_id_list(self, sentry_unit, process_name,
                            expect_success=True):
        """Get a list of process ID(s) from a single sentry juju unit
        for a single process name.

        :param sentry_unit: Amulet sentry instance (juju unit)
        :param process_name: Process name
        :param expect_success: If False, expect the PID to be missing,
            raise if it is present.
        :returns: List of process IDs
        """
        cmd = 'pidof -x "{}"'.format(process_name)
        if not expect_success:
            cmd += " || exit 0 && exit 1"
        output, code = sentry_unit.run(cmd)
        if code != 0:
            msg = ('{} `{}` returned {} '
                   '{}'.format(sentry_unit.info['unit_name'],
                               cmd, code, output))
            amulet.raise_status(amulet.FAIL, msg=msg)
        return str(output).split()

    def get_unit_process_ids(self, unit_processes, expect_success=True):
        """Construct a dict containing unit sentries, process names, and
        process IDs.

        :param unit_processes: A dictionary of Amulet sentry instance
            to list of process names.
        :param expect_success: if False expect the processes to not be
            running, raise if they are.
        :returns: Dictionary of Amulet sentry instance to dictionary
            of process names to PIDs.
        """
        pid_dict = {}
        for sentry_unit, process_list in six.iteritems(unit_processes):
            pid_dict[sentry_unit] = {}
            for process in process_list:
                pids = self.get_process_id_list(
                    sentry_unit, process, expect_success=expect_success)
                pid_dict[sentry_unit].update({process: pids})
        return pid_dict

    def validate_unit_process_ids(self, expected, actual):
        """Validate process id quantities for services on units."""
        self.log.debug('Checking units for running processes...')
        self.log.debug('Expected PIDs: {}'.format(expected))
        self.log.debug('Actual PIDs: {}'.format(actual))

        if len(actual) != len(expected):
            return ('Unit count mismatch.  expected, actual: {}, '
                    '{} '.format(len(expected), len(actual)))

        for (e_sentry, e_proc_names) in six.iteritems(expected):
            e_sentry_name = e_sentry.info['unit_name']
            if e_sentry in actual.keys():
                a_proc_names = actual[e_sentry]
            else:
                return ('Expected sentry ({}) not found in actual dict data.'
                        '{}'.format(e_sentry_name, e_sentry))

            if len(e_proc_names.keys()) != len(a_proc_names.keys()):
                return ('Process name count mismatch.  expected, actual: {}, '
                        '{}'.format(len(expected), len(actual)))

            for (e_proc_name, e_pids), (a_proc_name, a_pids) in \
                    zip(e_proc_names.items(), a_proc_names.items()):
                if e_proc_name != a_proc_name:
                    return ('Process name mismatch.  expected, actual: {}, '
                            '{}'.format(e_proc_name, a_proc_name))

                a_pids_length = len(a_pids)
                fail_msg = ('PID count mismatch. {} ({}) expected, actual: '
                            '{}, {} ({})'.format(e_sentry_name, e_proc_name,
                                                 e_pids, a_pids_length,
                                                 a_pids))

                # If expected is a list, ensure at least one PID quantity match
                if isinstance(e_pids, list) and \
                        a_pids_length not in e_pids:
                    return fail_msg
                # If expected is not bool and not list,
                # ensure PID quantities match
                elif not isinstance(e_pids, bool) and \
                        not isinstance(e_pids, list) and \
                        a_pids_length != e_pids:
                    return fail_msg
                # If expected is bool True, ensure 1 or more PIDs exist
                elif isinstance(e_pids, bool) and \
                        e_pids is True and a_pids_length < 1:
                    return fail_msg
                # If expected is bool False, ensure 0 PIDs exist
                elif isinstance(e_pids, bool) and \
                        e_pids is False and a_pids_length != 0:
                    return fail_msg
                else:
                    self.log.debug('PID check OK: {} {} {}: '
                                   '{}'.format(e_sentry_name, e_proc_name,
                                               e_pids, a_pids))
        return None

    def validate_list_of_identical_dicts(self, list_of_dicts):
        """Check that all dicts within a list are identical."""
        hashes = []
        for _dict in list_of_dicts:
            hashes.append(hash(frozenset(_dict.items())))

        self.log.debug('Hashes: {}'.format(hashes))
        if len(set(hashes)) == 1:
            self.log.debug('Dicts within list are identical')
        else:
            return 'Dicts within list are not identical'

        return None

    def validate_sectionless_conf(self, file_contents, expected):
        """A crude conf parser.  Useful to inspect configuration files which
        do not have section headers (as would be necessary in order to use
        the configparser).  Such as openstack-dashboard or rabbitmq confs."""
        for line in file_contents.split('\n'):
            if '=' in line:
                args = line.split('=')
                if len(args) <= 1:
                    continue
                key = args[0].strip()
                value = args[1].strip()
                if key in expected.keys():
                    if expected[key] != value:
                        msg = ('Config mismatch.  Expected, actual:  {}, '
                               '{}'.format(expected[key], value))
                        amulet.raise_status(amulet.FAIL, msg=msg)

    def get_unit_hostnames(self, units):
        """Return a dict of juju unit names to hostnames."""
        host_names = {}
        for unit in units:
            host_names[unit.info['unit_name']] = \
                str(unit.file_contents('/etc/hostname').strip())
        self.log.debug('Unit host names: {}'.format(host_names))
        return host_names

    def run_cmd_unit(self, sentry_unit, cmd):
        """Run a command on a unit, return the output and exit code."""
        output, code = sentry_unit.run(cmd)
        if code == 0:
            self.log.debug('{} `{}` command returned {} '
                           '(OK)'.format(sentry_unit.info['unit_name'],
                                         cmd, code))
        else:
            msg = ('{} `{}` command returned {} '
                   '{}'.format(sentry_unit.info['unit_name'],
                               cmd, code, output))
            amulet.raise_status(amulet.FAIL, msg=msg)
        return str(output), code

    def file_exists_on_unit(self, sentry_unit, file_name):
        """Check if a file exists on a unit."""
        try:
            sentry_unit.file_stat(file_name)
            return True
        except IOError:
            return False
        except Exception as e:
            msg = 'Error checking file {}: {}'.format(file_name, e)
            amulet.raise_status(amulet.FAIL, msg=msg)

    def file_contents_safe(self, sentry_unit, file_name,
                           max_wait=60, fatal=False):
        """Get file contents from a sentry unit.  Wrap amulet file_contents
        with retry logic to address races where a file checks as existing,
        but no longer exists by the time file_contents is called.
        Return None if file not found. Optionally raise if fatal is True."""
        unit_name = sentry_unit.info['unit_name']
        file_contents = False
        tries = 0
        while not file_contents and tries < (max_wait / 4):
            try:
                file_contents = sentry_unit.file_contents(file_name)
            except IOError:
                self.log.debug('Attempt {} to open file {} from {} '
                               'failed'.format(tries, file_name,
                                               unit_name))
                time.sleep(4)
                tries += 1

        if file_contents:
            return file_contents
        elif not fatal:
            return None
        elif fatal:
            msg = 'Failed to get file contents from unit.'
            amulet.raise_status(amulet.FAIL, msg)

    def port_knock_tcp(self, host="localhost", port=22, timeout=15):
        """Open a TCP socket to check for a listening sevice on a host.

        :param host: host name or IP address, default to localhost
        :param port: TCP port number, default to 22
        :param timeout: Connect timeout, default to 15 seconds
        :returns: True if successful, False if connect failed
        """

        # Resolve host name if possible
        try:
            connect_host = socket.gethostbyname(host)
            host_human = "{} ({})".format(connect_host, host)
        except socket.error as e:
            self.log.warn('Unable to resolve address: '
                          '{} ({}) Trying anyway!'.format(host, e))
            connect_host = host
            host_human = connect_host

        # Attempt socket connection
        try:
            knock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            knock.settimeout(timeout)
            knock.connect((connect_host, port))
            knock.close()
            self.log.debug('Socket connect OK for host '
                           '{} on port {}.'.format(host_human, port))
            return True
        except socket.error as e:
            self.log.debug('Socket connect FAIL for'
                           ' {} port {} ({})'.format(host_human, port, e))
            return False

    def port_knock_units(self, sentry_units, port=22,
                         timeout=15, expect_success=True):
        """Open a TCP socket to check for a listening sevice on each
        listed juju unit.

        :param sentry_units: list of sentry unit pointers
        :param port: TCP port number, default to 22
        :param timeout: Connect timeout, default to 15 seconds
        :expect_success: True by default, set False to invert logic
        :returns: None if successful, Failure message otherwise
        """
        for unit in sentry_units:
            host = unit.info['public-address']
            connected = self.port_knock_tcp(host, port, timeout)
            if not connected and expect_success:
                return 'Socket connect failed.'
            elif connected and not expect_success:
                return 'Socket connected unexpectedly.'

    def get_uuid_epoch_stamp(self):
        """Returns a stamp string based on uuid4 and epoch time.  Useful in
        generating test messages which need to be unique-ish."""
        return '[{}-{}]'.format(uuid.uuid4(), time.time())

    # amulet juju action helpers:
    def run_action(self, unit_sentry, action,
                   _check_output=subprocess.check_output,
                   params=None):
        """Translate to amulet's built in run_action(). Deprecated.

        Run the named action on a given unit sentry.

        params a dict of parameters to use
        _check_output parameter is no longer used

        @return action_id.
        """
        self.log.warn('charmhelpers.contrib.amulet.utils.run_action has been '
                      'deprecated for amulet.run_action')
        return unit_sentry.run_action(action, action_args=params)

    def wait_on_action(self, action_id, _check_output=subprocess.check_output):
        """Wait for a given action, returning if it completed or not.

        action_id a string action uuid
        _check_output parameter is no longer used
        """
        data = amulet.actions.get_action_output(action_id, full_output=True)
        return data.get(u"status") == "completed"

    def status_get(self, unit):
        """Return the current service status of this unit."""
        raw_status, return_code = unit.run(
            "status-get --format=json --include-data")
        if return_code != 0:
            return ("unknown", "")
        status = json.loads(raw_status)
        return (status["status"], status["message"])



from testtools import TestCase
import tempfile
import shutil
import subprocess
import six
import mock

from os.path import exists, join, isdir

from charmhelpers.contrib.ssl import service


class ServiceCATest(TestCase):

    def setUp(self):
        super(ServiceCATest, self).setUp()
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        super(ServiceCATest, self).tearDown()
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    @mock.patch("charmhelpers.contrib.ssl.service.log")
    def test_init(self, *args):
        """
        Tests that a ServiceCA is initialized with the correct directory
        layout.
        """
        ca_root_dir = join(self.temp_dir, 'ca')
        ca = service.ServiceCA('fake-name', ca_root_dir)
        ca.init()

        paths_to_verify = [
            'certs/',
            'crl/',
            'newcerts/',
            'private/',
            'private/cacert.key',
            'cacert.pem',
            'serial',
            'index.txt',
            'ca.cnf',
            'signing.cnf',
        ]

        for path in paths_to_verify:
            full_path = join(ca_root_dir, path)
            self.assertTrue(exists(full_path),
                            'Path {} does not exist'.format(full_path))

            if path.endswith('/'):
                self.assertTrue(isdir(full_path),
                                'Path {} is not a dir'.format(full_path))

    @mock.patch("charmhelpers.contrib.ssl.service.log")
    def test_create_cert(self, *args):
        """
        Tests that a generated certificate is valid against the ca.
        """
        ca_root_dir = join(self.temp_dir, 'ca')
        ca = service.ServiceCA('fake-name', ca_root_dir)
        ca.init()

        ca.get_or_create_cert('fake-cert')

        # Verify that the cert belongs to the ca
        self.assertTrue('fake-cert' in ca)

        full_cert_path = join(ca_root_dir, 'certs', 'fake-cert.crt')
        cmd = ['openssl', 'verify', '-verbose',
               '-CAfile', join(ca_root_dir, 'cacert.pem'), full_cert_path]

        output = subprocess.check_output(cmd,
                                         stderr=subprocess.STDOUT).strip()
        expected = '{}: OK'.format(full_cert_path)
        if six.PY3:
            expected = bytes(expected, 'utf-8')
        self.assertEqual(expected, output)
